# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for performing bulk queries on the mod database and repositories
"""

import logging
import multiprocessing
import os
import re
import shutil
import sys
from collections import defaultdict
from logging import info
from multiprocessing import Process
from time import sleep
from typing import (
    AbstractSet,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from progressbar import AnimatedMarker, ProgressBar, UnknownLength, Variable

from portmodlib.atom import Atom, FQAtom, QualifiedAtom, atom_sat
from portmodlib.colour import blue, bright, cyan, green, lblue, lgreen, red, yellow
from portmodlib.l10n import l10n
from portmodlib.portmod import Group, PackageIndexData, Person
from portmodlib.portmod import query as native_query
from portmodlib.portmod import update_index as native_update_index
from portmodlib.usestr import parse_usestr

from .config import get_config
from .config.mask import is_masked
from .config.use import get_use, get_use_expand, use_reduce
from .download import get_total_download_size
from .globals import env
from .loader import load_all, load_all_installed, load_installed_pkg, load_pkg
from .prefix import get_prefixes
from .pybuild import Pybuild
from .repo import get_repo
from .repo.metadata import get_global_use
from .repo.metadata import get_package_metadata as get_native_metadata
from .repo.metadata import get_use_expand_values

Maintainer = Union[Group, Person]
Maintainers = Union[Maintainer, List[Maintainer]]


def get_maintainer_strings(maintainers: Maintainers) -> List[str]:
    if not isinstance(maintainers, list):
        maintainers = [maintainers]
    return [str(maintainer) for maintainer in maintainers]


def get_maintainer_string(maintainers: Maintainers) -> str:
    return list_maintainers_to_human_strings(get_maintainer_strings(maintainers))


def list_maintainers_to_human_strings(maintainers: List[str]) -> str:
    """return the list of maintainers as a human readible string"""
    result = ""
    for maintainer_id in range(len(maintainers)):
        maintainer = maintainers[maintainer_id]
        if maintainer_id >= len(maintainers) - 1:  # the last
            result += maintainer
        elif maintainer_id >= len(maintainers) - 2:  # the second last
            result += maintainer + " and "
        else:
            result += maintainer + ", "
    return result


def print_depgraph(
    mod: Pybuild, level: int, level_limit: int, seen: Set[Pybuild]
) -> int:
    """
    Recursively prints the dependency graph for the given mod.

    Note that use conditionals and or statements are ignored.
    This prints out all possible dependencies, not actual dependencies.
    """
    if level > level_limit:
        return level - 1

    deps = parse_usestr(mod.DEPEND + " " + mod.RDEPEND, token_class=Atom)

    if mod in seen and deps:
        print(" " * (level + 2) + "-- " + l10n("omit-already-displayed-tree"))
        return level - 1
    max_level = level

    if isinstance(seen, frozenset):
        seen = set()
    seen.add(mod)

    def print_token(token, conditionals=None):
        atom = Atom(token)
        mod = max(load_pkg(atom), key=lambda pkg: pkg.version)
        enabled, _ = get_use(mod)

        def colour(flag):
            if flag.rstrip("?").lstrip("-!") in enabled:
                return bright(red(flag))
            return bright(blue(flag))

        use = map(colour, atom.USE)
        use_str = ""
        if atom.USE:
            use_str = f'[{" ".join(use)}]'

        if conditionals:
            dep = f"( {' '.join(conditionals)} ( {atom.strip_use()} ) ) "
        else:
            dep = f"({atom.strip_use()}) "

        print(" " * (level + 1) + f"-- {bright(green(mod.ATOM.CPF))} " + dep + use_str)
        return print_depgraph(mod, level + 1, level_limit, seen)

    def print_deps(deps, conditionals: List):
        nonlocal max_level
        if isinstance(deps, list):
            if not deps:
                return
            if isinstance(deps[0], str) and deps[0] == "||":
                for inner_token in deps[1:]:
                    print_deps(inner_token, conditionals)
            elif isinstance(deps[0], str) and deps[0].endswith("?"):
                for inner_token in deps[1:]:
                    print_deps(inner_token, conditionals + [deps[0]])
            else:
                for inner_token in deps:
                    print_deps(inner_token, conditionals)
        else:
            max_level = max(max_level, print_token(deps, conditionals))

    print_deps(deps, [])

    return max_level


def str_strip(value: str) -> str:
    return re.sub("(( +- +)|(:))", "", value)


def str_squelch_sep(value: str) -> str:
    return re.sub(r"[-_\s]+", " ", value)


def query(
    value: str,
    limit: int = 10,
) -> List[PackageIndexData]:
    """
    Finds mods that contain the given value in the given field
    """
    if not env.PREFIX_NAME:
        raise RuntimeError("Queries must be done inside a prefix!")
    return native_query(env.INDEX, env.PREFIX_NAME, value, limit)


# FIXME: For this to be indexed, we'll need to add DEPEND and RDEPEND to the index
# This also means indexing individual package versions as well as aggregated data
def query_depends(atom: Atom, all_mods=False) -> List[Tuple[FQAtom, str]]:
    """
    Finds mods that depend on the given atom
    """
    if all_mods:
        mods = load_all()
    else:
        mods = load_all_installed()

    depends = []
    for mod in mods:
        if not all_mods:
            enabled, disabled = get_use(mod)
            atoms = use_reduce(
                mod.RDEPEND + " " + mod.DEPEND,
                enabled,
                disabled,
                token_class=Atom,
                flat=True,
            )
        else:
            atoms = use_reduce(
                mod.RDEPEND + " " + mod.DEPEND,
                token_class=Atom,
                matchall=True,
                flat=True,
            )

        for dep_atom in atoms:
            if dep_atom != "||" and atom_sat(dep_atom, atom):
                depends.append((mod.ATOM, dep_atom))
    return depends


def get_flag_string(
    name: Optional[str],
    enabled: Iterable[str],
    disabled: Iterable[str],
    installed: Optional[AbstractSet[str]] = None,
    *,
    verbose: bool = True,
    display_minuses=True,
):
    """
    Displays flag configuration

    Enabled flags are displayed as blue
    If the installed flag list is passed, flags that differ from the
    installed set will be green
    if name is None, the name prefix will be omitted and no quotes will
    surround the flags
    """

    def disable(string: str) -> str:
        if display_minuses:
            return "-" + string
        return string

    flags = []
    for flag in sorted(enabled):
        if installed is not None and flag not in installed:
            flags.append(bright(lgreen(flag)))
        elif verbose:
            flags.append(red(bright(flag)))

    for flag in sorted(disabled):
        if installed is not None and flag in installed:
            flags.append(bright(lgreen(disable(flag))))
        elif verbose:
            if display_minuses:
                flags.append(blue(disable(flag)))
            else:
                flags.append(lblue(disable(flag)))

    inner = " ".join(flags)

    if not flags:
        return None

    if name:
        return f'{name}="{inner}"'

    return inner


def display_search_results(
    packages: Sequence[PackageIndexData],
    summarize: bool = True,
    numbers: bool = False,
    file=sys.stdout,
):
    """
    Prettily formats a list of package metadata for use in search results
    """
    count = 0

    footnotes: Dict[str, str] = {}
    footnotenum = "a"

    def footnote(repo: str):
        nonlocal footnotenum
        if repo not in footnotes:
            footnotes[repo] = str(footnotenum)
            footnotenum = chr(ord(footnotenum) + 1)
        return lblue(f"[{footnotes[repo]}]")

    for package in packages:
        count += 1
        group = load_pkg(Atom(package.cpn))
        sortedmods = sorted(group, key=lambda pkg: pkg.version)
        newest = sortedmods[-1]
        installed = load_installed_pkg(Atom(package.cpn))
        download = get_total_download_size([newest])

        if installed is not None:
            installed_str = blue(bright(installed.PVR)) + footnote(installed.REPO)

            flags = {flag.lstrip("+") for flag in installed.IUSE if "_" not in flag}
            usestr = get_flag_string(
                None, installed.INSTALLED_USE & flags, flags - installed.INSTALLED_USE
            )
            texture_options = {
                size
                for mod in group
                for size in use_reduce(
                    installed.TEXTURE_SIZES, matchall=True, flat=True
                )
            }
            texture = next(
                (
                    use.lstrip("texture_size_")
                    for use in installed.INSTALLED_USE
                    if use.startswith("texture_size_")
                ),
                None,
            )
            if isinstance(texture, str):
                texture_string = get_flag_string(
                    "TEXTURE_SIZE", [texture], texture_options - {texture}
                )
            else:
                texture_string = None
            use_expand_strings = []
            for use in get_config().get("USE_EXPAND", []):
                if use in get_config().get("USE_EXPAND_HIDDEN", []):
                    continue
                enabled_expand, disabled_expand = get_use_expand(installed, use)
                if enabled_expand or disabled_expand:
                    string = get_flag_string(use, enabled_expand, disabled_expand, None)
                    use_expand_strings.append(string)

            all_flags: List[str] = list(
                filter(None, [usestr, texture_string] + use_expand_strings)
            )
            if all_flags:
                installed_str += " {" + " ".join(all_flags) + "}"
        else:
            installed_str = "not installed"

        # List of version numbers, prefixed by either (~) or ** depending on
        # keyword for user's arch. Followed by use flags, including use expand
        version_str = ""
        versions = []
        ARCH = env.prefix().ARCH
        for mod in sortedmods:
            if mod.INSTALLED:
                continue

            if ARCH in mod.KEYWORDS:
                version_nofmt = mod.PVR
                version = green(mod.PVR)
            elif "~" + ARCH in mod.KEYWORDS:
                version_nofmt = "(~)" + mod.PVR
                version = yellow(version_nofmt)
            else:
                version_nofmt = "**" + mod.PVR
                version = red(version_nofmt)

            repo = mod.REPO

            if is_masked(mod.ATOM, mod.REPO):
                versions.append(red("[M]" + version_nofmt) + footnote(repo))
            else:
                versions.append(version + footnote(repo))
        version_str = " ".join(versions)
        flags = {
            flag.lstrip("+") for mod in group for flag in mod.IUSE if "_" not in flag
        }
        usestr = get_flag_string(None, [], flags, display_minuses=False)
        texture_options = {
            size
            for mod in group
            for size in use_reduce(mod.TEXTURE_SIZES, matchall=True, flat=True)
        }
        texture_string = get_flag_string(
            "TEXTURE_SIZE", [], texture_options, display_minuses=False
        )
        use_expand_strings = []
        for use in get_config().get("USE_EXPAND", []):
            if use in get_config().get("USE_EXPAND_HIDDEN", []):
                continue
            flags = {
                re.sub(f"^{use.lower()}_", "", flag)
                for flag in mod.IUSE_EFFECTIVE
                for mod in group
                if flag.startswith(f"{use.lower()}_")
            }
            if flags:
                string = get_flag_string(use, [], flags, None, display_minuses=False)
                use_expand_strings.append(string)

        all_flags = list(filter(None, [usestr, texture_string] + use_expand_strings))
        if all_flags:
            version_str += " {" + " ".join(all_flags) + "}"

        # If there are multiple URLs, remove any formatting from the pybuild and
        # add padding

        if logging.root.level <= logging.DEBUG:
            if package.homepage:
                formatted_homepages = "\n                 ".join(
                    [package.homepage] + package.other_homepages
                )

                homepage_str = (
                    f"       {green(l10n('package-homepage'))} {formatted_homepages}"
                )
            print(
                f"{cyan(count) if numbers else ''} {green('*')}  {bright(package.cpn)}",
                f"       {green(l10n('package-name'))} {package.name}",
                f"       {green(l10n('package-available-versions'))} {version_str}",
                f"       {green(l10n('package-installed-version'))} {installed_str}",
                f"       {green(l10n('package-size-of-files'))} {download}",
                sep=os.linesep,
                file=file,
            )
            if package.homepage:
                print(homepage_str, file=file)
            print(
                f"       {green(l10n('package-description'))} {str_squelch_sep(newest.DESC)}",
                f"       {green(l10n('package-license'))} {newest.LICENSE}",
                sep=os.linesep,
                file=file,
            )
            if package.upstream_maintainers:
                print(
                    f"       {green(l10n('package-upstream-author'))} {list_maintainers_to_human_strings(package.upstream_maintainers)}",
                    file=file,
                )
            print(file=file)
        else:
            print(
                f"{cyan(count) if numbers else ''} {green('*')}  {bright(package.cpn)}",
                f"       {green(l10n('package-available-versions'))} {version_str}",
                f"       {green(l10n('package-installed-version'))} {installed_str}",
                sep=os.linesep,
                file=file,
            )
            if package.homepage:
                print(
                    f"       {green(l10n('package-homepage'))} {package.homepage}",
                    file=file,
                )
            print(
                f"       {green(l10n('package-description'))} {str_squelch_sep(newest.DESC)}",
                sep=os.linesep,
                file=file,
            )

    if footnotes:
        print()
        for repo in footnotes:
            print(footnote(repo), f'"{repo}"', get_repo(repo).location)

    if summarize:
        print("\n" + l10n("packages-found", num=len(packages)), file=file)


class FlagDesc:
    """Use flag descriptions"""

    def __init__(self, desc: str):
        self.desc = desc

    def __str__(self):
        return self.desc


class LocalFlagDesc(FlagDesc):
    """Local use flag description"""

    def __init__(self, pkg: Pybuild, desc: str):
        super().__init__(desc)
        self.pkg = pkg

    def __repr__(self):
        return f"LocalDesc({self.pkg}, {self.desc})"


class UseExpandDesc(FlagDesc):
    """Local use flag description"""

    def __init__(self, category: str, flag: str, desc: str):
        super().__init__(desc)
        self.flag = flag
        self.category = category

    def __repr__(self):
        return f"UseExpandDesc({self.category}, {self.desc})"


def get_flag_desc(pkg: Pybuild, flag: str) -> Optional[FlagDesc]:
    """Returns the description for the given use flag"""
    repo_root = get_repo(pkg.REPO).location

    global_use = get_global_use(repo_root)
    metadata = get_native_metadata(pkg)

    if metadata and flag in metadata.use:
        return LocalFlagDesc(pkg, metadata.use[flag])
    if flag in global_use:
        return FlagDesc(global_use[flag])
    if flag.startswith("texture_size_"):
        size = flag.replace("texture_size_", "")
        return UseExpandDesc("texture_size", size, l10n("texture-size-desc", size=size))
    if "_" in flag:  # USE_EXPAND
        use_expand = flag.rsplit("_", 1)[0]
        suffix = flag.replace(use_expand + "_", "")
        use_expand_desc = get_use_expand_values(repo_root, use_expand).get(suffix)
        if use_expand_desc:
            return UseExpandDesc(use_expand, suffix, use_expand_desc)

    return None


def get_flags(
    pkg: Pybuild,
) -> Tuple[Dict[str, str], Dict[str, str], Dict[str, Dict[str, str]]]:
    """
    Returns all use flags and their descriptions for the given package

    returns:
        Three dictionaries, one each for local flags, global flags and use_expand flags,
        in that order. The use expand flags are subdivided for each use_expand category.
    """
    repo_root = get_repo(pkg.REPO).location

    global_use = get_global_use(repo_root)
    metadata = get_native_metadata(pkg)

    local_flags = {}
    global_flags = {}
    use_expand_flags: DefaultDict[str, Dict[str, str]] = defaultdict(dict)

    for flag in pkg.IUSE_EFFECTIVE:
        if metadata and flag in metadata.use:
            local_flags[flag] = metadata.use[flag]
        elif flag in global_use:
            global_flags[flag] = global_use[flag]
        elif flag.startswith("texture_size_"):
            size = flag.replace("texture_size_", "")
            desc = l10n("texture-size-desc", size=size)
            use_expand_flags["texture_size"][size] = desc
        elif "_" in flag:  # USE_EXPAND
            use_expand = flag.rsplit("_", 1)[0]
            suffix = flag.replace(use_expand + "_", "")
            use_expand_desc = get_use_expand_values(repo_root, use_expand).get(suffix)
            if use_expand_desc:
                use_expand_flags[use_expand][suffix] = use_expand_desc
        else:
            # No description Found.
            # Might be an installed package without metadata.yaml
            continue

    return local_flags, global_flags, use_expand_flags


def get_package_metadata(package: QualifiedAtom) -> Optional[PackageIndexData]:
    packages = load_pkg(package)
    if not packages:
        return None

    pkgs = sorted(packages, key=lambda x: x.ATOM.version)
    newest = pkgs[-1]

    assert env.PREFIX_NAME, "A prefix must be specified"

    pkg_metadata = PackageIndexData(
        cpn=package.CPN,
        prefix=env.PREFIX_NAME,
        category=package.C,
        package=package.PN,
        name=newest.NAME,
        desc=newest.DESC,
    )

    homepages = use_reduce(newest.HOMEPAGE, matchall=True, flat=True)

    if homepages:
        pkg_metadata.homepage = homepages[0]
    if len(homepages) > 1:
        pkg_metadata.other_homepages = homepages[1:]

    if newest.LICENSE:
        pkg_metadata.license = newest.LICENSE

    for pkg in pkgs:
        metadata = get_native_metadata(pkg)
        if metadata:
            pkg_metadata.tags |= metadata.tags
            if metadata.longdescription:
                pkg_metadata.longdescription = metadata.longdescription
            if metadata.maintainer:
                pkg_metadata.maintainers = get_maintainer_strings(metadata.maintainer)
            if metadata.upstream:
                if metadata.upstream.maintainer:
                    pkg_metadata.upstream_maintainers = get_maintainer_strings(
                        metadata.upstream.maintainer
                    )
                if metadata.upstream.doc:
                    pkg_metadata.upstream_doc = metadata.upstream.doc
                if metadata.upstream.bugs_to:
                    pkg_metadata.upstream_bugs_to = metadata.upstream.bugs_to
                if metadata.upstream.changelog:
                    pkg_metadata.upstream_changelog = metadata.upstream.changelog

    return pkg_metadata


def _get_index_data() -> List[PackageIndexData]:
    if not env.PREFIX_NAME:
        return []

    def get_package_names():
        seen = set()
        for pkg in load_all():
            if pkg.CPN not in seen:
                seen.add(pkg.CPN)
                yield pkg.CPN
        for pkg in load_all_installed():
            if pkg.CPN not in seen:
                seen.add(pkg.CPN)
                yield pkg.CPN

    info("Beginning index update for prefix " + env.PREFIX_NAME)
    if sys.stderr.isatty() or env.TESTING:
        try:
            from progressbar import GranularBar as Bar
        except ImportError:
            from progressbar import Bar
        from progressbar import ETA, BouncingBar, Counter, ProgressBar, Timer, Variable

        package_names = set()
        bar = ProgressBar(
            redirect_stdout=True,
            widgets=[
                Variable("status", format="Collecting packages"),
                BouncingBar(),
                Counter(),
                " ",
                Timer(),
            ],
        )

        i = 0
        bar.start()
        for name in get_package_names():
            package_names.add(name)
            bar.update(i)
            i += 1
        bar.finish()

        package_data = []
        bar = ProgressBar(
            redirect_stdout=True,
            widgets=[
                Variable("status", format="Loading package data"),
                Bar(),
                Counter(),
                " ",
                ETA(),
            ],
            max_value=len(package_names),
        )
        bar.start()
        i = 0
        for cpn in package_names:
            metadata = get_package_metadata(cpn)
            if metadata:
                package_data.append(metadata)
            bar.update(i)
            i += 1
        bar.finish()
    else:
        package_data = []
        for name in get_package_names():
            metadata = get_package_metadata(name)
            if metadata:
                package_data.append(metadata)
    return package_data


def _run_progress_bar(pipe):
    bar = ProgressBar(
        widgets=[
            Variable("status", format="{variables.status}"),
            " ",
            AnimatedMarker(),
        ],
        max_value=UnknownLength,
        variables={"status": "Updating index"},
    )
    bar.start()
    while not pipe.poll():
        bar.update()
        sleep(0.1)
    bar.update(status="Done updating index.")
    bar.finish()


def _commit_index(package_data: List[PackageIndexData]):
    os.makedirs(env.INDEX, exist_ok=True)

    if sys.stderr.isatty() or env.TESTING:
        pipe, child_pipe = multiprocessing.Pipe()
        process = Process(target=_run_progress_bar, args=(child_pipe,))
        process.start()
        native_update_index(env.INDEX, package_data)
        pipe.send(True)
        process.join()
    else:
        info("Updating index...")
        native_update_index(env.INDEX, package_data)
        info("Done updating index.")


def update_index():
    orig_prefix = env.PREFIX_NAME
    if os.path.exists(env.INDEX):
        shutil.rmtree(env.INDEX)
    package_data = []
    if env.TESTING:
        package_data.extend(_get_index_data())
    else:
        for prefix in get_prefixes():
            env.set_prefix(prefix)
            package_data.extend(_get_index_data())
    _commit_index(package_data)
    env.set_prefix(orig_prefix)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
CLI for interacting with individual pybuild files
"""

import os
import py_compile
import sys
from glob import glob
from logging import error, info, warning

from portmod.download import download_source
from portmod.globals import env
from portmod.loader import load_file
from portmod.parsers.manifest import FileType, Manifest, ManifestEntry
from portmod.pybuild import Pybuild, manifest_path
from portmod.repo import get_repo_root
from portmod.repo.metadata import (
    is_pybuild_version_banned,
    is_pybuild_version_deprecated,
)
from portmod.repos import Repo
from portmod.source import HashAlg
from portmodlib.l10n import l10n


def create_manifest(pkg: Pybuild):
    """
    Automatically downloads mod DIST files (if not already in cache)
    and creates a manifest file
    """
    manifest = Manifest(manifest_path(pkg.FILE))

    existing_sources = set()

    # Collect the names of existing files
    for file in glob(os.path.join(os.path.dirname(pkg.FILE), "*.pybuild")):
        if file == pkg.FILE:
            continue
        thismod = load_file(file)
        existing_sources |= {
            source.name for source in thismod.get_sources(matchall=True)
        }

    # Remove files not referenced by any of the package files
    # This will also remove files, if any, referenced by the package file being added to the manifest
    for name in {name for name in manifest.entries if name not in existing_sources}:
        del manifest.entries[name]

    # Add sources from the package file to manifest
    for source in load_file(pkg.FILE).get_sources(matchall=True):
        filename = download_source(pkg, source)
        if filename is None:
            error("Unable to get shasum for unavailable file " + source.name)
            continue

        manifest.add_entry(
            ManifestEntry.from_path(
                FileType.DIST, filename, source.name, [HashAlg.BLAKE3, HashAlg.MD5]
            )
        )

    # Write changes to manifest
    manifest.write()


def pybuild_validate(file_name):
    # Verify that pybuild is valid python
    py_compile.compile(file_name, doraise=True)

    # Verify fields of pybuild
    pkg = load_file(file_name)
    if is_pybuild_version_banned(get_repo_root(file_name), pkg._PYBUILD_VER):
        raise TypeError(f"{file_name} uses Pybuild{pkg._PYBUILD_VER}, which is banned")
    if is_pybuild_version_deprecated(get_repo_root(file_name), pkg._PYBUILD_VER):
        warning(f"{file_name} uses Pybuild{pkg._PYBUILD_VER}, which is deprecated")
    pkg.validate()


def pybuild_manifest(file_name):
    if not os.path.exists(file_name):
        raise FileNotFoundError(l10n("file-does-not-exist", file=file_name))

    repo_root = get_repo_root(file_name)

    if repo_root is None:
        raise FileNotFoundError(l10n("repository-does-not-exist"))

    # Register repo in case it's not already in repos.cfg
    REAL_ROOT = os.path.realpath(repo_root)
    if not any([REAL_ROOT == os.path.realpath(repo.location) for repo in env.REPOS]):
        sys.path.append(os.path.join(repo_root))
        env.REPOS.append(Repo(os.path.basename(repo_root), repo_root))

    if env.ALLOW_LOAD_ERROR:
        raise Exception("Cannot allow load errors when generating manifest!")

    mod = load_file(file_name)

    create_manifest(mod)
    info(l10n("created-manifest", atom=mod.ATOM))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import pytest

from portmod._deps import resolve
from portmod.config.sets import add_set, get_set, remove_set
from portmod.config.use import add_global_use
from portmod.loader import load_installed_pkg
from portmod.merge import configure
from portmod.transactions import Downgrade, New, Reinstall, Update
from portmodlib.atom import Atom, Version

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_simple(setup):
    """Tests that simple dependency resolution works"""
    selected = {Atom("test/test")}
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert transactions.pkgs[0].pkg.CPN == "test/test"
    assert isinstance(transactions.pkgs[0], New)

    configure(["test/test"])
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert transactions.pkgs[0].pkg.CPN == "test/test"
    assert isinstance(transactions.pkgs[0], Reinstall)


def test_rebuild(setup):
    """
    Tests that packages are selected to be rebuilt, even if we don't
    use the Category-PackageName format
    """
    selected = {Atom("~test/test-1.0")}
    configure(selected)
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert transactions.pkgs[0].pkg.CPN == "test/test"
    assert transactions.pkgs[0].pkg.REPO == "test"
    assert isinstance(transactions.pkgs[0], Reinstall)


def test_upgrade(setup):
    """Tests that upgrades resolve correctly"""
    selected = {Atom("test/test")}
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert transactions.pkgs[0].pkg.version > Version("1.0")
    assert isinstance(transactions.pkgs[0], Update)


def test_oneshot(setup):
    """Tests that oneshot resolves correctly"""
    selected = {Atom("test/test")}
    configure(selected)
    transactions = resolve(selected, set(), selected, set(), set())
    assert len(transactions.pkgs) == 1
    assert not transactions.pkgs[0].pkg.version > Version("2.0")
    assert not Version("2.0") > transactions.pkgs[0].pkg.version
    assert Version("2.0") == transactions.pkgs[0].pkg.version
    assert isinstance(transactions.pkgs[0], Reinstall)
    assert not transactions.new_selected


def test_downgrade(setup):
    """Tests that downgrades resolve correctly"""
    configure(["=test/test-2.0"])
    selected = {Atom("=test/test-1.0")}
    transactions = resolve(selected, set(), selected, selected, set())
    assert len(transactions.pkgs) == 1
    assert Version("2.0") > transactions.pkgs[0].pkg.version
    assert isinstance(transactions.pkgs[0], Downgrade)


def test_auto_depclean(setup):
    """
    Tests that auto depclean doesn't change configuration to remove packages

    There are two possible changes that it shouldn't make for this test.
    test7-1.0 could be downgraded to test7-0.1 to remove the dependencies
    And test7-1.0's flags could be disabled to remove the dependencies.
    """
    configure(["=test/test7-1.0[baz]"])
    pkg = load_installed_pkg(Atom("test/test5"))
    assert pkg
    transactions = resolve(
        enabled={Atom("test/test7")},
        disabled=set(),
        explicit={Atom("test/test7")},
        selected=set(),
        selected_sets={"world"},
        update=True,
        deep=True,
        depclean=True,
    )
    assert len(transactions.pkgs) == 0


def test_upgrade_configuration(setup):
    """
    Tests that upgrades will pull in new packages and change configuration if necessary
    """
    configure(["@installed"], delete=True)
    configure(
        ["=test/test7-0.1"],
        update=True,
        deep=True,
    )
    add_global_use("baz")
    pkg = load_installed_pkg(Atom("test/test5"))
    assert not pkg
    transactions = resolve(
        enabled=get_set("world"),
        disabled=set(),
        explicit=get_set("world"),
        selected=set(),
        selected_sets={"world"},
        update=True,
        deep=True,
    )
    for change in transactions.config:
        raise Exception(f"Unexpected configuration change {change}")
    assert len(transactions.pkgs) == 3
    for change in transactions.pkgs:
        if isinstance(change, Update):
            assert change.pkg.P == "test7-1.0"
        elif isinstance(change, New):
            assert change.pkg.PN in {"test5", "test4"}
        else:
            raise Exception(f"Unexpected transaction ({change.REPR}) {change.pkg}")

    configure(["test/test7"], depclean=True)


def test_nondeep_depth(setup):
    """
    Tests that, when not using deep mode, packages which are in the world set but
    have not been installed will pull in all their dependencies
    """
    add_set("selected-packages", Atom("=test/test6-1.0"))
    transactions = resolve(
        enabled=set(),
        disabled=set(),
        explicit=set(),
        selected=set(),
        selected_sets=set(),
        update=True,
    )
    # test6 depends on test3 which depends on test4
    packages = ["test6", "test3", "test4"]
    for change in transactions.pkgs:
        assert isinstance(change, New)
        assert change.pkg.PN in packages
        packages.remove(change.pkg.PN)

    assert not packages
    remove_set("selected-packages", Atom("=test/test6-1.0"))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import pytest

from portmod.repo.keywords import accepts

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_stable_visibility():
    assert accepts(["test"], ["test"])
    assert not accepts(["test"], ["test2"])
    assert accepts(["~test"], ["test"])
    assert accepts(["~*"], ["test"])
    assert accepts(["*"], ["test"])


def test_wild_stable_visibility():
    assert accepts(["test"], ["*", "test2"])
    assert accepts(["~test"], ["*", "test2"])
    assert accepts(["test"], ["*", "test"])
    assert accepts(["~test"], ["*", "test"])


def test_wild_testing_visibility():
    assert accepts(["~test"], ["~*", "test2"])
    assert accepts(["~test"], ["~*", "~test2"])
    assert not accepts(["test"], ["~*", "test2"])
    assert not accepts(["test"], ["~*", "~test2"])


def test_testing_visibility():
    assert not accepts(["test"], ["~test"])
    assert accepts(["~test"], ["~test"])
    assert accepts(["~*"], ["~test"])
    assert not accepts(["*"], ["~test"])


def test_wild_visibility():
    assert accepts(["test"], ["**"])
    assert accepts(["~test"], ["**"])
    assert accepts([], ["**"])


def test_untested_visibility():
    assert not accepts(["test"], [])
    assert not accepts(["~test"], [])
    assert accepts(["**"], [])



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests for use flag aliases
"""

import pytest

from portmod.config.use import remove_use
from portmod.globals import env
from portmod.loader import SandboxedError, load_installed_pkg
from portmod.merge import configure
from portmod.repos import get_repos
from portmodlib.atom import Atom

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """sets up and tears down test environment"""
    dictionary = setup_env("test")
    env.REPOS = get_repos()
    yield dictionary
    tear_down_env()


def test_simple_alias(setup):
    """Tests that use flag aliases match the installed state of the aliased package"""
    configure(["test/test2"])

    pkg = load_installed_pkg(Atom("test/test2"))
    assert pkg
    assert "alias1" not in pkg.INSTALLED_USE
    assert "alias2" not in pkg.INSTALLED_USE

    configure(
        ["test/test2", "test/test7"],
        deep=True,
        update=True,
        verbose=True,
    )

    pkg = load_installed_pkg(Atom("test/test2"))
    assert pkg
    assert "alias1" in pkg.INSTALLED_USE
    assert "alias2" not in pkg.INSTALLED_USE

    configure(["test/test2", "test/test7"], delete=True)


def test_alias_with_use_flag(setup):
    """Tests that use flag aliases with use flag dependencies behave correctly"""
    configure(["test/test2"])

    pkg = load_installed_pkg(Atom("test/test2"))
    assert pkg
    assert "alias1" not in pkg.INSTALLED_USE
    assert "alias2" not in pkg.INSTALLED_USE

    configure(
        ["test/test2", "test/test4"],
        deep=True,
        update=True,
    )

    pkg = load_installed_pkg(Atom("test/test2"))
    assert pkg
    assert "alias1" not in pkg.INSTALLED_USE
    assert "alias2" not in pkg.INSTALLED_USE

    configure(
        ["test/test2", "test/test4[foo]"],
        deep=True,
        update=True,
    )

    pkg = load_installed_pkg(Atom("test/test2"))
    assert pkg
    assert "alias1" not in pkg.INSTALLED_USE
    assert "alias2" in pkg.INSTALLED_USE

    configure(["test/test2", "test/test4"], delete=True)
    remove_use("foo", Atom("test/test4"))


def test_alias_nofetch(setup):
    """Tests that use flag aliases work correctly in the scope of pkg_nofetch"""
    # Should raise an exception if
    with pytest.raises(SandboxedError):
        configure(
            ["test/test-nofetch", "test/test4[foo]"],
            deep=True,
            update=True,
        )



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests some otherwise untested parts of the interface
"""

import io
import os
import sys

import pytest

from portmod._cli.main import main
from portmod.globals import env
from portmod.loader import load_installed_pkg
from portmodlib.atom import Atom

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    yield setup_env("test")
    tear_down_env()


def test_validate():
    """Tests that validate works correctly"""
    sys.argv = ["portmod", "test", "merge", "test", "test2", "--no-confirm"]
    main()
    sys.argv = ["portmod", "test", "validate"]
    main()


def test_sync():
    """Tests that portmod sync works correctly"""
    sys.argv = ["portmod", "sync"]
    main()


def test_info():
    """Tests that portmod info works correctly"""
    sys.argv = ["portmod", "test", "info"]
    with pytest.raises(SystemExit):
        main()


def test_search(monkeypatch):
    """Tests that portmod search works correctly"""
    sys.argv = ["portmod", "test", "search", "test"]
    monkeypatch.setattr("sys.stdin", io.StringIO("\n"))
    main()


def test_use():
    """Tests that portmod use works correctly"""
    sys.argv = ["portmod", "test", "use", "-E", "foo"]
    main()
    sys.argv = ["portmod", "test", "merge", "test4", "--no-confirm"]
    main()
    pkg = load_installed_pkg(Atom("test/test4"))
    assert pkg
    assert "foo" in pkg.get_use()


def test_mirror():
    """Tests that portmod --version works correctly"""
    sys.argv = ["portmod", "mirror", os.path.join(env.TMP_DIR, "mirror"), "test"]
    main()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from typing import Any, List, NamedTuple, TypeVar, Union

from portmodlib.atom import Atom, FQAtom
from portmodlib.usestr import parse_usestr

_T = TypeVar("_T", bound="Token")


class Token(NamedTuple):
    """A formula token"""

    value: str
    polarity: bool = True

    def __str__(self) -> str:
        if self.polarity:
            return self.value
        else:
            return "-" + self.value

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Token):
            # Note: order is important to avoid unnecessary string comparisions
            return self.polarity == other.polarity and self.value == other.value
        return False

    def neg(self: _T) -> _T:
        return self.__class__(self.value, not self.polarity)


class AtomToken(Token):
    """Token descirbing a package"""

    value: FQAtom


class FlagToken(Token):
    """Token describing a use flag dependency on a package"""

    value: FQAtom

    def __str__(self) -> str:
        if self.polarity:
            return list(self.value.USE)[0]
        else:
            return "-" + list(self.value.USE)[0]


class VariableToken(Token):
    """Token for a variable which exists only in the context of the formula"""


def token_conflicts(token1: Token, token2: Token) -> bool:
    """
    Returns true if and only if two tokens, which use minus-format (e.g. -foo)
    to indicate a disabled token, conflict. E.g. foo and -foo
    """
    return token1.value == token2.value and token1.polarity != token2.polarity


def expand_use_conditionals(tokens: List[str]) -> List[Union[str, List]]:
    """Expands any conditional use dependencies in the token tree"""
    result = []
    for token in tokens:
        if isinstance(token, list):
            result.append(expand_use_conditionals(token))
        elif isinstance(token, Atom) and token.USE:
            for flag in token.USE:
                stripped = token.strip_use()
                sflag = flag.rstrip("?=").lstrip("!")
                if flag.endswith("?") and not flag.startswith("!"):
                    result += parse_usestr(
                        f"{sflag}? ( {stripped}[{sflag}] ) !{sflag}? ( {stripped} )",
                        Atom,
                    )
                elif flag.endswith("?") and flag.startswith("!"):
                    result += parse_usestr(
                        f"{sflag}? ( {stripped} ) !{sflag}? ( {stripped}[-{sflag}] )",
                        Atom,
                    )
                elif flag.endswith("=") and not flag.startswith("!"):
                    result += parse_usestr(
                        f"{sflag}? ( {stripped}[{sflag}] ) !{sflag}? ( {stripped}[-{sflag}] )",
                        Atom,
                    )
                elif flag.endswith("=") and flag.startswith("!"):
                    result += parse_usestr(
                        f"{sflag}? ( {stripped}[-{sflag}] ) !{sflag}? ( {stripped}[{sflag}] )",
                        Atom,
                    )
                else:
                    result.append(Atom(stripped + f"[{flag}]"))
        else:
            result.append(token)
    return result



# Copyright 2022 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import sys
from io import StringIO

import pytest

from portmod._cli.main import main
from portmod.cfg_protect import get_redirections
from portmod.globals import env
from portmod.merge import configure

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    os.makedirs(env.prefix().CONFIG_PROTECT_DIR, exist_ok=True)
    yield dictionary
    tear_down_env()


def test_cfg_protect(setup):
    """
    Tests that file protected by CFG_PROTECT don't get their changes overwritten
    when the package is re-installed
    """
    configure(["=test/test-1.0-r1"])
    path = os.path.join(env.prefix().ROOT, "etc", "test")

    with open(path, "a", encoding="utf-8") as file:
        print("bar = baz", file=file)
    with open(path, encoding="utf-8") as file:
        contents = file.readlines()

    configure(["=test/test-1.0-r1"])

    with open(path, encoding="utf-8") as file:
        assert contents == file.readlines()

    # Since the file was the same when re-installed,
    # it shouldn't get added to the list of pending config file updates
    assert not list(get_redirections())

    configure(["=test/test-1.0-r1"], delete=True)
    configure
    os.remove(path)


def test_cfg_protect_changed(setup, monkeypatch):
    """
    Tests that files protected by CFG_PROTECT get installed as separate files
    and registered with cfg_protect when an update changes the file
    """
    configure(["=test/test-1.0-r1"])
    path = os.path.join(env.prefix().ROOT, "etc", "test")

    with open(path, "a", encoding="utf-8") as file:
        print("bar = baz", file=file)
    with open(path, encoding="utf-8") as file:
        contents = file.readlines()

    configure(["=test/test-2.0"])

    # File should not have been changed directly
    with open(path, encoding="utf-8") as file:
        assert contents == file.readlines()

    # Since the file was the same when re-installed,
    # it shouldn't get added to the list of pending config file updates
    redirections = list(get_redirections())
    assert redirections
    src, dst = redirections[0]
    assert dst == path
    assert src == path + ".__cfg_protect__"
    with open(src, encoding="utf-8") as file:
        assert file.read() == "foo = baz\n"

    env.INTERACTIVE = True
    monkeypatch.setattr("sys.stdin", StringIO("y\n"))
    sys.argv = ["portmod", "test", "cfg-update"]
    main()

    with open(path, encoding="utf-8") as file:
        assert file.read() == "foo = baz\n"
    env.INTERACTIVE = False


def is_empty(path: str) -> bool:
    if not os.path.exists(path):
        return True

    with open(path, "r") as file:
        if file.read():
            return False

    return True


def test_cfg_protect_license(setup):
    """
    Tests that cfg_protect protects the package.accept_license file
    if installing in non-interactive mode
    """

    env.TESTING = False
    # Normally cfg_protect on configuration files is not enabled in testing mode
    assert is_empty(
        os.path.join(env.prefix().CONFIG_DIR, "package.accept_license")
    ), "package.accept_license should not exist at the start"

    try:
        configure(["test/test-eula"])
    # Configure will fail since license changes must be accepted
    except SystemExit:
        pass

    assert is_empty(
        os.path.join(env.prefix().CONFIG_DIR, "package.accept_license")
    ), "package.accept_license should not contain anything after running the merge"

    assert os.path.exists(
        os.path.join(env.prefix().CONFIG_DIR, "package.accept_license.__cfg_protect__")
    )

    env.TESTING = True


def test_cfg_protect_keywords(setup):
    """
    Tests that cfg_protect protects the package.accept_keywords file
    if installing in non-interactive mode
    """

    env.TESTING = False
    # Normally cfg_protect on configuration files is not enabled in testing mode
    assert is_empty(
        os.path.join(env.prefix().CONFIG_DIR, "package.accept_keywords")
    ), "package.accept_keywords should not exist at the start"

    try:
        configure(["=test/test0-2.0_alpha"])
    # Configure will fail since keyword changes must be accepted
    except SystemExit:
        pass

    assert is_empty(
        os.path.join(env.prefix().CONFIG_DIR, "package.accept_keywords")
    ), "package.accept_keywords should not contain anything after running the merge"

    assert os.path.exists(
        os.path.join(env.prefix().CONFIG_DIR, "package.accept_keywords.__cfg_protect__")
    )

    env.TESTING = True


def test_cfg_protect_use(setup):
    """
    Tests that cfg_protect protects the package.use file
    if installing in non-interactive mode
    """

    env.TESTING = False
    # Normally cfg_protect on configuration files is not enabled in testing mode
    assert is_empty(
        os.path.join(env.prefix().CONFIG_DIR, "package.use")
    ), "package.use should not exist at the start"

    try:
        configure(["=test/test-1.0-r2[test1]"])
    # Configure will fail since use changes must be accepted
    except SystemExit:
        pass

    assert is_empty(
        os.path.join(env.prefix().CONFIG_DIR, "package.use")
    ), "package.use should not contain anything after running the merge"

    assert os.path.exists(
        os.path.join(env.prefix().CONFIG_DIR, "package.use.__cfg_protect__")
    )

    env.TESTING = True



# Copyright 2022 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from typing import Generator, Iterable, List, Optional, Set, Tuple, Union

from portmodlib.globals import local_mods, root
from portmodlib.pybuild import BasePybuild
from portmodlib.usestr import check_required_use


class File:
    """
    Represents important installed files and their metadata

    .. warning::
        This class has been deprecated as of Portmod 2.4.
        It will be removed in Portmod 3.0
    """

    def __init__(
        self,
        NAME: str,
        REQUIRED_USE: str = "",
        OVERRIDES: Union[str, List[str]] = [],
        **kwargs,
    ):
        """
        File objects also support a REQUIRED_USE variable, for , and an OVERRIDES variable for overriding other plugins in the load order.
        """
        self.__keys__: Set[str] = set()
        self.NAME: str = NAME
        """Name of the file relative to the root of the InstallDir"""
        self.REQUIRED_USE: str = REQUIRED_USE
        """
        Requirements for installing this file

        The default empty string is always satisfied.
        See Pybuild2.REQUIRED_USE for details on the syntax.
        """
        self.OVERRIDES: Union[str, List[str]] = OVERRIDES
        """
        A list of files which this overrides when sorting (if applicable).

        Can either be in the form of a string containing use-conditionals (note that
        this does not support files that contain spaces) or a list of files to override.
        Note that these overridden files are not considered masters and do not need to
        be present.

        For archives it determines the order in which the fallback archives will be
        searched during VFS lookups.
        """
        if REQUIRED_USE:
            self._add_kwarg("REQUIRED_USE", REQUIRED_USE)
        if OVERRIDES:
            self._add_kwarg("OVERRIDES", OVERRIDES)

        for key in kwargs:
            self._add_kwarg(key, kwargs[key])

    def _add_kwarg(self, key, value):
        self.__dict__[key] = value
        self.__keys__.add(key)

    def __repr__(self):
        return self.__str__()

    def __str__(self) -> str:
        kvps = []
        for key in self.__keys__:
            value = getattr(self, key)
            if isinstance(value, str):
                kvps.append(f'{key}="{getattr(self, key)}"')
            else:
                kvps.append(f"{key}={getattr(self, key)}")

        separator = ""
        if kvps:
            separator = ", "
        return f'File("{self.NAME}"' + separator + ", ".join(kvps) + ")"

    def _to_cache(self):
        cache = {"NAME": self.NAME}
        for key in self.__keys__:
            cache[key] = getattr(self, key)

        cache["__type__"] = "File"
        return cache


class InstallDir:
    """
    Represents a directory in the Virtual File System

    Note that arbitrary arguments can be passed to the constructor, as
    repositories may make use of custom information.
    See the repository-level documentation for such information.

    .. warning::
        This class has been deprecated as of Portmod 2.4.
        It will be removed in Portmod 3.0
    """

    def __init__(
        self,
        PATH: str,
        REQUIRED_USE: str = "",
        PATCHDIR: str = ".",
        S: Optional[str] = None,
        WHITELIST: Optional[List[str]] = None,
        BLACKLIST: Optional[List[str]] = None,
        RENAME: Optional[str] = None,
        DATA_OVERRIDES: str = "",
        ARCHIVES: Iterable[File] = (),
        VFS: Optional[bool] = None,
        DOC: Iterable[str] = (),
        **kwargs,
    ):
        self.PATH: str = PATH
        """
        The path to the data directory that this InstallDir represents
        relative to the root of the archive it is contained within.
        """
        self.REQUIRED_USE: str = REQUIRED_USE
        """
        A list of use flags with the same format as the package's
        REQUIRED_USE variable which enable the InstallDir if satisfied.
        Defaults to an empty string that is always satisfied.
        """
        self.PATCHDIR: str = PATCHDIR
        """
        The destination path of the InstallDir within the package's directory.

        Defaults to ".", i.e. the root of the mod directory. If multiple InstallDirs
        share the same PATCHDIR they will be installed into the same directory in the
        order that they are defined in the INSTALL_DIRS list.
        Each unique PATCHDIR has its own entry in the VFS, and its own sorting rules
        """
        self.S: Optional[str] = S
        """
        The source directory corresponding to this InstallDir.

        Similar function to S for the entire pybuild, this determines which directory
        contains this InstallDir, and generally corresponds to the name of the source
        archive, minus extensions. This is required for packages that contain more
        than one source, but is automatically detected for those with only one source
        if it is not specified, and will first take the value of Pybuild2.S, then the
        source's file name without extension if the former was not defined.
        """
        self.WHITELIST: Optional[List[str]] = WHITELIST
        """
        If present, only installs files matching the patterns in this list.
        fnmatch-style globbing patterns (e.g. * and [a-z]) can be used
        """
        self.BLACKLIST: Optional[List[str]] = BLACKLIST
        """
        If present, does not install files matching the patterns in this list.
        fnmatch-style globbing patterns (e.g. * and [a-z]) can be used
        """
        self.RENAME: Optional[str] = RENAME
        """
        Destination path of this directory within the final directory.

        E.g.::

            InstallDir("foo/bar", PATCHDIR=".", RENAME="bar")

        Will install the contents of ``foo/bar`` (in the source) into the directory
        ``bar`` inside the package's installation directory (and also the VFS).
        """
        self.DATA_OVERRIDES: str = DATA_OVERRIDES
        """
        A list of packages that this InstallDir should override in the VFS

        This only has a different effect from Pybuild1.DATA_OVERRIDES if multiple PATCHDIRs
        are set, as it can define overrides for individual PATCHDIRS, while
        Pybuild1.DATA_OVERRIDES affects all PATCHDIRs.
        See Pybuild1.DATA_OVERRIDES for details of the syntax.
        """
        self.ARCHIVES: List[File] = list(ARCHIVES)
        """
        A list of File objects representing VFS archives.

        These will be searched, in order, during VFS file lookups if the file is not
        present in the package directories.
        """
        self.VFS: Optional[bool] = VFS
        """
        Whether or not this InstallDir gets added to the VFS

        Defaults to the value of the VFS variable in the profile configuration
        """
        self.DOC: List[str] = list(DOC)
        """
        A list of patterns matching documentation files within the package

        This documentation will be installed separately
        fnmatch-style globbing patterns (e.g. * and [a-z]) can be used.
        """
        self.__keys__: Set[str] = set()
        if ARCHIVES:
            self._add_kwarg("ARCHIVES", ARCHIVES)
        if PATCHDIR != ".":
            self._add_kwarg("PATCHDIR", PATCHDIR)
        for arg in [
            "DATA_OVERRIDES",
            "RENAME",
            "BLACKLIST",
            "WHITELIST",
            "S",
            "REQUIRED_USE",
        ]:
            if getattr(self, arg):
                self._add_kwarg(arg, getattr(self, arg))
        for key in kwargs:
            self._add_kwarg(key, kwargs[key])

    def _add_kwarg(self, key, value):
        if isinstance(value, list):
            new_value = []
            for item in value:
                if isinstance(item, dict) and item.get("__type__") == "File":
                    file = dict(item)
                    file.pop("__type__", None)
                    new_value.append(File(**file))
                else:
                    new_value.append(item)
            value = new_value

        self.__dict__[key] = value
        self.__keys__.add(key)

    def get_files(self):
        """Generator function yielding file subattributes of the installdir"""
        for key in self.__dict__:
            if isinstance(getattr(self, key), list):
                for item in getattr(self, key):
                    if isinstance(item, File):
                        yield item

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        kvps = []
        for key in self.__keys__:
            value = getattr(self, key)
            if isinstance(value, str):
                kvps.append(f'{key}="{getattr(self, key)}"')
            else:
                kvps.append(f"{key}={getattr(self, key)}")

        separator = ""
        if kvps:
            separator = ", "
        return f'InstallDir("{self.PATH}"' + separator + ", ".join(kvps) + ")"

    def _to_cache(self):
        cache = {"PATH": self.PATH}
        for key in self.__keys__:
            value = getattr(self, key)
            if isinstance(value, list):
                new = []
                for item in value:
                    if isinstance(item, File):
                        new.append(item._to_cache())
                    else:
                        new.append(item)
                value = new
            cache[key] = value

        return cache


def get_directories(package: BasePybuild) -> Generator[InstallDir, None, None]:
    """
    Returns all enabled InstallDir objects in INSTALL_DIRS
    """
    assert package._PYBUILD_VER == 1
    for install_dir in getattr(package, "INSTALL_DIRS"):
        if check_required_use(
            install_dir.REQUIRED_USE, package.get_use(), package.valid_use
        ):
            yield install_dir


def get_files(
    package: BasePybuild, typ: str
) -> Generator[Tuple[InstallDir, File], None, None]:
    """
    Returns all enabled files and their directories
    """
    for install_dir in get_directories(package):
        if hasattr(install_dir, typ):
            for file in getattr(install_dir, typ):
                if check_required_use(
                    file.REQUIRED_USE, package.get_use(), package.valid_use
                ):
                    yield install_dir, file


# Note: This function will only work inside the sandbox, as otherwise INSTALL_DEST
# won't be an environment variable
def _get_install_dir_dest(pkg: BasePybuild):
    install_dir_dest = os.environ.get("INSTALL_DEST", ".")
    for attr in dir(pkg):
        if not attr.startswith("_") and isinstance(getattr(pkg, attr), str):
            install_dir_dest = install_dir_dest.replace(
                "{" + attr + "}", getattr(pkg, attr)
            )
    return os.path.normpath(install_dir_dest)


def get_dir_path(pkg: BasePybuild, install_dir: InstallDir) -> str:
    """Returns the installed path of the given InstallDir"""
    # Local dirs should be relative to LOCAL_MODS
    if "local" in pkg.PROPERTIES:
        path = os.path.normpath(os.path.join(local_mods(), pkg.PN))
    else:
        path = os.path.normpath(
            os.path.join(root(), _get_install_dir_dest(pkg), install_dir.PATCHDIR)
        )
    if os.path.islink(path):
        return os.readlink(path)
    else:
        return path


def get_file_path(pkg: BasePybuild, install_dir: InstallDir, esp: File) -> str:
    return os.path.join(get_dir_path(pkg, install_dir), esp.NAME)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import re
from typing import Set

from .portmod import get_masters as native_get_masters


def get_masters(file: str) -> Set[str]:
    """
    Detects masters for the given file

    .. warning::
        This function has been deprecated as of Portmod 2.4
        it will be removed in Portmod 3.0

    args:
        file: File to be examined
    returns:
        A set of all the master names
    """
    _, ext = os.path.splitext(file)
    if re.match(r"\.(esp|esm|omwaddon|omwgame)", ext, re.IGNORECASE):
        return set(native_get_masters(file))
    return set()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

# Note: Some phase function descriptions have are derived from the
# Gentoo Package Manager Specification, version 7.
#   https://projects.gentoo.org/pms/7/pms.html
# This is licensed under the CC-BY-SA-3.0 License

import json
import lzma
import os
from pathlib import Path
from typing import Any, List, Optional, Set, Union

from .atom import Atom, FQAtom, QualifiedAtom, Version
from .source import Source
from .usestr import use_reduce


class BasePybuild:
    """
    Interface describing the Pybuild Type
    Only describes elements that are cached.
    This class cannot be used to install/uninstall mods
    """

    __file__ = __file__
    _PYBUILD_VER: int

    version: Version
    ATOM: FQAtom
    RDEPEND: str = ""
    DEPEND: str = ""
    SRC_URI: str = ""
    P: Atom
    PF: Atom
    PN: Atom
    CATEGORY: str
    PV: str
    PR: str
    PVR: str
    CPN: QualifiedAtom
    CP: QualifiedAtom
    REQUIRED_USE: str = ""
    REQUIRED_USE_EFFECTIVE: str = ""
    RESTRICT: str = ""
    PROPERTIES: str = ""
    IUSE_EFFECTIVE: Set[str] = set()
    IUSE: Set[str] = set()
    TEXTURE_SIZES: str = ""
    DESC: str = ""
    NAME: str = ""
    HOMEPAGE: str = ""
    LICENSE: str = ""
    KEYWORDS: str = ""
    FILE: str
    REPO: str
    INSTALLED: bool = False
    S: Optional[str] = None  # Primary directory during prepare and install operations:w
    PATCHES: str = ""
    # Phase functions defined by the pybuild (or superclasses other than Pybuild1/2)
    # Used to determine if a function should be run, as certain functions don't have any default
    # behaviour
    FUNCTIONS: List[str] = []
    # Only set in phase functions
    USE: Set[str]

    def get_use(self) -> Set[str]:
        """Returns the enabled use flags for the package"""
        return self.USE

    def valid_use(self, use: str) -> bool:
        """Returns true if the given flag is a valid use flag for this mod"""
        return use in self.IUSE_EFFECTIVE


class FullPybuild(BasePybuild):
    """Interface describing the Pybuild Type"""

    __file__ = __file__
    REPO_PATH: Optional[str]
    __pybuild__: str

    # Variables defined during the install/removal process
    A: List[Source]  # List of enabled sources
    D: str  # Destination directory where the mod is to be installed
    FILESDIR: str  # Path of the directory containing additional repository files
    ROOT: str  # Path of the installed directory of the mod
    T: str  # Path of temp directory
    UNFETCHED: List[Source]  # List of sources that need to be fetched
    USE: Set[str]  # Enabled use flags
    WORKDIR: str  # Path of the working directory

    # Note: declared as a string, but converted into a set during __init__
    IUSE: Union[Set[str], str] = ""  # type: ignore
    KEYWORDS = ""

    def __init__(self):
        self.FILE = self.__class__.__pybuild__

        category = Path(self.FILE).resolve().parent.parent.name
        # Note: type will be fixed later by the loader and will become an FQAtom
        self.ATOM = Atom(  # type: ignore
            "{}/{}".format(category, os.path.basename(self.FILE)[: -len(".pybuild")])
        )
        self.version = self.ATOM.version

        self.REPO_PATH = str(Path(self.FILE).resolve().parent.parent.parent)
        repo_name_path = os.path.join(self.REPO_PATH, "profiles", "repo_name")
        if os.path.exists(repo_name_path):
            with open(repo_name_path, "r") as repo_file:
                self.REPO = repo_file.readlines()[0].rstrip()
            self.ATOM = FQAtom("{}::{}".format(self.ATOM, self.REPO))

        self.P = Atom(self.ATOM.P)
        self.PN = Atom(self.ATOM.PN)
        self.PV = self.ATOM.PV
        self.PF = Atom(self.ATOM.PF)
        self.PR = self.ATOM.PR or "r0"
        self.CATEGORY = self.ATOM.C
        self.R = self.ATOM.R
        self.CP = QualifiedAtom(self.ATOM.CP)
        self.CPN = QualifiedAtom(self.ATOM.CPN)
        self.PVR = self.ATOM.PVR

        self.IUSE_EFFECTIVE = set()

        if isinstance(self.IUSE, str):
            self.IUSE = set(self.IUSE.split())
            self.IUSE_EFFECTIVE |= set([use.lstrip("+") for use in self.IUSE])
        else:
            raise TypeError("IUSE must be a space-separated list of use flags")

        if type(self.KEYWORDS) is str:
            self.KEYWORDS = set(self.KEYWORDS.split())  # type: ignore # pylint: disable=no-member
        else:
            raise TypeError("KEYWORDS must be a space-separated list of keywords")

        self.REQUIRED_USE_EFFECTIVE: str = self.REQUIRED_USE
        if type(self.TEXTURE_SIZES) is str:
            texture_sizes = use_reduce(self.TEXTURE_SIZES, matchall=True)
            texture_size_flags = [
                "texture_size_{}".format(size) for size in texture_sizes
            ]
            self.IUSE_EFFECTIVE |= set(texture_size_flags)
            if texture_size_flags:
                self.REQUIRED_USE_EFFECTIVE += (
                    " ^^ ( " + " ".join(texture_size_flags) + " )"
                )
        else:
            raise TypeError(
                "TEXTURE_SIZES must be a string containing a space separated list of "
                "texture sizes"
            )

    def pkg_nofetch(self):
        """
        The pkg_nofetch function is run when the fetch phase of an fetch-restricted
        package is run, and the relevant source files are not available. It should
        direct the user to download all relevant source files from their respective
        locations, with notes concerning licensing if applicable.

        ``pkg_nofetch`` must require no write access to any part of the filesystem.
        """

    def pkg_pretend(self):
        """
        May be used to carry out sanity checks early on in the install process

        Note that the default does nothing, and it will not even be executed unless defined.

        ``pkg_pretend`` is run separately from the main phase function sequence, and does not
        participate in any kind of environment saving. There is no guarantee that any of
        an package’s dependencies will be met at this stage, and no guarantee that the system
        state will not have changed substantially before the next phase is executed.

        ``pkg_pretend`` must not write to the filesystem and the initial working directory
        should not be expected to be consistent.
        """

    def src_unpack(self):
        """
        The ``src_unpack`` function extracts all of the package’s sources.

        The initial working directory must be ``self.WORKDIR``, and the default implementation
        used when the package lacks the src_unpack function shall behave as::

            self.unpack(self.A)
        """

    def src_prepare(self):
        """
        The src_prepare function can be used for post-unpack source preparation.

        The initial working directory is :py:attr:`Pybuild2.S`, falling back to
        :py:attr:`Pybuild2.WORKDIR` if the directory does not exist.

        The default implementation used when the package lacks
        the ``src_prepare`` function shall behave as::

            if self.PATCHES:
                for patch in use_reduce(self.PATCHES, self.USE, flat=True):
                    path = os.path.join(self.FILESDIR, patch)
                    apply_patch(path)
        """

    def src_install(self):
        """
        The src_install function installs the package’s content to a directory specified in :py:attr:`Pybuild2.D`.

        The initial working directory is :py:attr:`Pybuild2.S`, falling back to :py:attr:`Pybuild.WORKDIR`
        if the directory does not exist.

        The default implementation used when the package lacks the ``src_install`` function shall behave as::

            for pattern in self.DOCS:
                self.dodoc(pattern)
        """

    def pkg_prerm(self):
        """
        Function called immediately before package removal

        In Pybuild1, this function has full write permissions to ROOT.
        In Pybuild2 it only has read permissions.

        Note that the default does nothing, and it will not even be executed unless defined.
        """

    def pkg_postinst(self):
        """
        Function called immediately after package installation

        In Pybuild1, this function has full write permissions to ROOT.
        In Pybuild2 it only has read permissions.

        Note that the default does nothing, and it will not even be executed unless defined.
        """

    @staticmethod
    def execute(
        command: str, pipe_output: bool = False, pipe_error: bool = False
    ) -> Optional[str]:
        """Function pybuild files can use to execute native commands"""

    def validate(self) -> List[str]:
        """
        QA Checks pybuild structure. This has the same restrictions as code in the global
        package scope.

        returns:
            A list of error messages
        """
        return []


class FullInstalledPybuild(FullPybuild):
    """Interface describing the type of installed Pybuilds"""

    INSTALLED_USE: Set[str] = set()
    INSTALLED: bool = True
    INSTALLED_REBUILD_FILES: Optional[Any] = None


def get_installed_env(pkg: BasePybuild):
    environment = {}
    path = os.path.join(os.path.dirname(pkg.FILE), "environment.xz")

    if os.path.exists(path):
        environment_json = lzma.LZMAFile(path)
        try:
            environment = json.load(environment_json)
        except EOFError as e:
            raise RuntimeError(f"Failed to read {path}") from e
    return environment



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Depclean tests
"""

import pytest

from portmod.loader import load_all_installed, load_installed_pkg
from portmod.merge import configure, deselect
from portmodlib.atom import Atom

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_depclean(setup):
    """
    Tests that deselected mods are then depcleaned
    """
    configure(["test/test-1.0", "test/test2-1.0"])
    mod = load_installed_pkg(Atom("test/test2"))
    assert mod
    assert mod in load_all_installed()
    deselect(["test/test2"])
    configure([], depclean=True)
    assert not load_installed_pkg(Atom("test/test2"))

    mod = load_installed_pkg(Atom("test/test"))
    assert mod
    assert mod in load_all_installed()
    deselect(["test/test"])
    configure([], depclean=True)
    # Note: test/test is a system mod, so it cannot be removed


def test_noarg_depclean(setup):
    """
    Tests that deselected mods are then depcleaned
    """
    configure(["test/test6-1.0"])
    configure(["test/test6-1.0"], delete=True)
    assert load_installed_pkg(Atom("test/test3"))
    configure([], depclean=True)
    assert not load_installed_pkg(Atom("test/test3"))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Module for loading pybuilds within a sandboxed environment"""

import json
import logging
import os
import re
import shutil
import subprocess
import sys
import traceback
from functools import wraps
from logging import debug, error, warning
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    TypeVar,
    cast,
)

import portmod
from portmod.config import get_config, variable_data_dir
from portmod.execute import sandbox_execute
from portmod.functools import install_cache, prefix_aware_cache, system_cache
from portmod.vdb import vdb_path
from portmodlib.atom import Atom, FQAtom, atom_sat
from portmodlib.colour import green
from portmodlib.execute import execute
from portmodlib.l10n import l10n

from .cache import (
    PreviouslyEncounteredException,
    __load_mod_from_dict_cache,
    pybuild_dumper,
)
from .globals import env
from .perms import Permissions
from .pybuild import File, InstallDir, InstalledPybuild, Pybuild
from .repo.loader import (
    __safe_load_module,
    _iterate_installed,
    _iterate_pybuilds,
    find_installed_path,
    get_atom_from_path,
)


class AmbiguousAtom(Exception):
    """Indicates that multiple packages from different categories match"""

    def __init__(self, atom: Atom, packages: Iterable[Atom], fq: bool = False):
        message_id = "ambiguous-atom-fq" if fq else "ambiguous-atom"
        super().__init__(
            l10n(message_id, atom=green(atom))
            + "\n  "
            + green("\n  ".join(sorted(packages)))
        )


class SandboxedError(Exception):
    """Error raised when a sandboxed command fails"""

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return f"An error was encountered when execuring code inside the sandbox: {self.message}"


@system_cache
def _state_path(file) -> str:
    atom = get_atom_from_path(file)
    return os.path.join(env.TMP_DIR, atom.C, atom.P, "state")


def _delete_state(file):
    if os.path.exists(_state_path(file)):
        shutil.rmtree(_state_path(file))


def get_wrapper_code():
    # Preserve virtualenv
    venv_activate = ""
    if "VIRTUAL_ENV" in os.environ:
        if sys.platform == "win32":
            file = os.path.join(
                os.environ["VIRTUAL_ENV"], "scripts", "activate_this.py"
            )
        else:
            file = os.path.join(os.environ["VIRTUAL_ENV"], "bin", "activate_this.py")
        venv_activate = (
            'exec(open(r"' + file + '").read(), {"__file__": "' + file + '"})'
        )

    return f"""
import sys
from os import path as osp

{venv_activate}
if __name__ == "__main__":
    # Ignore the -c argument, and this code, which are the first two arguments passed to python
    del sys.argv[0:1]
    # Third argument should be a file within the portmod module.
    # This also takes the place of the program name in the argument list so that argparse
    # handles the remaining arguments correctly
    if osp.isfile(
        osp.join(
            osp.dirname(osp.dirname(osp.realpath(sys.argv[0]))), ".portmod_not_installed"
        )
    ):
        sys.path.insert(0, osp.dirname(osp.dirname(osp.realpath(sys.argv[0]))))

    from portmodlib._wrapper import main

    main()
"""


def load_module(file: str, state):
    from portmod.modules import Module, ModuleFunction

    module_data = __safe_load_module(file, False, None)

    perms = Permissions(
        rw_paths=[state.CACHE, state.ROOT],
        ro_paths=[
            # Builtin sets and the package database can't be modified by modules
            os.path.join(variable_data_dir(), "sets", "world"),
            os.path.join(variable_data_dir(), "sets", "world_sets"),
            os.path.join(variable_data_dir(), "db"),
            os.path.join(variable_data_dir(), "news"),
        ],
        global_read=True,
        tmp=state.TEMP,
    )

    def get_func_wrapper(function_name: str):

        if function_name == "update":

            def func_wrapper(state):
                return _sandbox_execute_module(
                    file,
                    function=function_name,
                    init=state.__dict__,
                    permissions=perms,
                )

        else:

            def func_wrapper(state, args):  # type: ignore
                return _sandbox_execute_module(
                    file,
                    function=function_name,
                    args=args,
                    init=state.__dict__,
                    permissions=perms,
                )

        return func_wrapper

    do_functions = {}
    descriptions = {}
    describe_options = {}
    describe_parameters = {}
    for globname in module_data:
        if globname.startswith("do_"):
            name = re.sub("^do_", "", globname)
            do_functions[name] = module_data[globname]

        match = re.match("^describe_(.*)$", globname)
        if match:
            descriptions[match.group(1)] = str(module_data[globname]())
        match = re.match("^describe_(.*)_options$", globname)
        if match:
            describe_options[match.group(1)] = module_data[globname]()
        match = re.match("^describe_(.*)_parameters$", globname)
        if match:
            describe_parameters[match.group(1)] = module_data[globname]()

    functions = []
    for function_name in do_functions:
        functions.append(
            ModuleFunction(
                function_name,
                descriptions.get(function_name),
                get_func_wrapper(function_name),
                describe_options.get(function_name),
                describe_parameters.get(function_name),
                state,
            )
        )
    module_name = os.path.basename(file)
    module_name, _ = os.path.splitext(module_name)

    return Module(
        module_name,
        module_data.get("__doc__"),
        sorted(functions, key=lambda x: x.name),
        state,
    )


@system_cache
def _get_library_dirs(path: str) -> Set[str]:
    """
    Returns the directories containing the libraries
    used by the executable at the given path

    Used to attempt to detect non-standard library directories
    """
    # Note: while we could just use direct paths of the libraries, this gives a little more
    # Flexibility for providing other libraries in the sandbox
    paths = set()

    if sys.platform == "win32":
        pass
    else:
        try:
            if sys.platform == "darwin":
                lines = execute(f"otool -L {path}", pipe_output=True, pipe_error=True)
            else:
                lines = execute(f"ldd {path}", pipe_output=True, pipe_error=True)
        except subprocess.CalledProcessError as err:
            if "not a dynamic executable" in str(err.stderr):
                return set()
            error(err.stderr.decode("utf-8"))

        # Add anything that looks like an absolute path
        for token in (lines or "").split():
            directory = None

            if token.startswith("/"):
                directory = token
            elif token.startswith("@executable_path/"):
                directory = os.path.normpath(
                    os.path.join(path, token.replace("@executable_path/", ""))
                )

            if directory:
                paths.add(os.path.dirname(directory))
    return paths


def _sandbox_execute(
    file_type: str,
    file: str,
    action: str,
    permissions: Permissions,
    *,
    save_state: bool = False,
    init: Optional[Dict[str, Any]] = None,
    curdir: Optional[str] = None,
    installed: bool = False,
    args: Optional[Any] = None,
    function: Optional[str] = None,
):
    assert file_type in ("pybuild", "module")
    python = os.path.realpath(sys.executable)

    abspath = os.path.abspath(file)
    old_curdir = os.getcwd()
    if not os.path.exists(curdir or env.TMP_DIR):
        os.makedirs(curdir or env.TMP_DIR)
    os.chdir(curdir or env.TMP_DIR)
    ro_paths = (
        set(permissions.ro_paths)
        | {env.prefix().ROOT, env.CONFIG_DIR, env.TMP_VDB, os.path.dirname(abspath)}
        | {repo.location for repo in env.REPOS}
        # Python site packages directories, etc.
        | set(sys.path)
    )

    # Binary search paths
    # Relative paths are ignored,
    # as the current directory may not be what it initially was
    splitchar = ";" if sys.platform == "win32" else ":"
    for path in os.environ["PATH"].split(splitchar):
        if os.path.isabs(path):
            ro_paths.add(path)

    if not curdir:
        ro_paths.add(env.TMP_DIR)
    # Detect Libraries used by executables in case of non-standard library locations
    for executable in ["python", "git", "bsatool"]:
        exec_path = shutil.which(executable)
        if exec_path:
            ro_paths |= _get_library_dirs(os.path.realpath(exec_path))
    command = [
        python,
        "-c",
        get_wrapper_code(),
        portmod.__file__,
        "--verbosity",
        logging.getLevelName(logging.root.level),
        file_type,
        abspath,
        action,
    ]
    if file_type == "pybuild":
        ro_paths.add(vdb_path())
    rw_paths = set(permissions.rw_paths)
    rw_paths.add(env.PYBUILD_TMP_DIR)
    if save_state:
        command += ["--state-dir", _state_path(file)]
        os.makedirs(_state_path(file), exist_ok=True)
        rw_paths.add(_state_path(file))
    if init:
        # Note: quotes must be escaped
        state_string = json.dumps(init, default=pybuild_dumper)
        command += ["--initial-state", state_string]
    if args:
        command += ["--args", json.dumps(args)]
    if function:
        command += ["--module-func", function]
    try:
        result = sandbox_execute(
            command,
            Permissions(
                permissions,
                ro_paths=sorted(ro_paths),
                rw_paths=sorted(rw_paths),
            ),
        )
    except subprocess.CalledProcessError as err:
        debug(err)
        raise SandboxedError(
            l10n("command-failed", path=abspath, command=action)
        ) from None
    finally:
        os.chdir(old_curdir)
    return result


def _sandbox_execute_module(
    file: str,
    *,
    permissions: Permissions = Permissions(),
    args: Optional[Any] = None,
    function: Optional[str] = None,
    init: Optional[Dict[str, Any]] = None,
):
    """
    Modules, as they are only executed after installation, have greater permissions than
    pybuilds. They have read-only access to the entire filesystem, though not to the
    network by default.

    Write access should be done using the CONFIG_PROTECT_DIR. There is a create_file
    function which can be used to create files in the CONFIG_PROTECT_DIR that shadow
    another file in the filesystem. The user will be prompted to overwrite the shadowed
    file when the module is finished executing.
    """
    # TODO: Allow modules to request network permissions
    os.makedirs(env.prefix().CONFIG_PROTECT_DIR, exist_ok=True)
    return _sandbox_execute(
        "module",
        file,
        "execute",
        Permissions(
            permissions,
            rw_paths=[env.prefix().CONFIG_PROTECT_DIR],
        ),
        init=init,
        function=function,
        args=args,
    )


def _sandbox_execute_pybuild(
    file: str,
    action: str,
    permissions: Permissions,
    *,
    save_state: bool = False,
    init: Optional[Dict[str, Any]] = None,
    curdir: Optional[str] = None,
    installed: bool = False,
):
    os.makedirs(env.WARNINGS_DIR, exist_ok=True)
    os.makedirs(env.MESSAGES_DIR, exist_ok=True)
    result = _sandbox_execute(
        "pybuild",
        file,
        action,
        permissions,
        save_state=save_state,
        init=init,
        curdir=curdir,
        installed=installed,
    )
    if action == "can-update-live":
        return result is True


V = TypeVar("V", bound=Any)


def __safe_load(user_function: Callable[..., V]) -> Callable[..., Optional[V]]:
    """
    Decorator that makes a function return None if it would otherwise raise an exception
    """

    @wraps(user_function)
    def decorating_function(name, *args, **kwargs):
        try:
            return user_function(name, *args, **kwargs)
        except PreviouslyEncounteredException as e:
            if env.ALLOW_LOAD_ERROR:
                return None
            raise e.previous
        except Exception as e:
            warning(e)
            if logging.root.level <= logging.DEBUG:
                traceback.print_exc()
            warning(l10n("could-not-load-pybuild", file=name))
            if env.ALLOW_LOAD_ERROR:
                return None
            raise e

    return cast(Callable[..., Optional[V]], decorating_function)


@install_cache
def load_installed_pkg(atom: Atom) -> Optional[InstalledPybuild]:
    """Loads packages from the installed database"""
    path = find_installed_path(atom)

    if path is not None:
        pkg = cast(Optional[InstalledPybuild], safe_load_file(path, installed=True))
        if pkg and atom_sat(pkg.ATOM, atom, ignore_name=True):
            return pkg

    if not atom.C or atom.C == "local":
        local_path = os.path.join(env.prefix().LOCAL_MODS, atom.PN)
        if os.path.exists(local_path):
            return _load_local_pkg(local_path)

    return None


@prefix_aware_cache
def load_pkg_fq(atom: FQAtom) -> Pybuild:
    """
    Loads package matching fully qualified atom.

    except:
        FileNotFoundError: If the package cannot be found
        AmbiguousAtom: If multiple packages match the given atom
    """
    assert isinstance(atom, FQAtom)
    if atom.R.endswith("::installed") or atom.R == "installed":
        installed = load_installed_pkg(atom)
        if installed:
            return installed

        raise FileNotFoundError(l10n("not-found", atom=atom))

    packages: List[Pybuild] = []
    for file in _iterate_pybuilds(atom, atom.R):
        pkg = safe_load_file(file)
        if pkg is None:
            continue

        packages.append(pkg)

    if len(packages) > 1:
        raise AmbiguousAtom(atom, [pkg.ATOM for pkg in packages], fq=True)
    if len(packages) == 1:
        return packages[0]

    raise FileNotFoundError(l10n("not-found", atom=atom))


@install_cache
def load_pkg(atom: Atom, *, repo_name: Optional[str] = None) -> List[Pybuild]:
    """
    Loads all mods matching the given atom
    There may be multiple versions in different repos,
    as well versions with different version or release numbers

    :param atom: Mod atom to load.
    :param repo_name: If present, the name of the repository tree to search.
                      The masters of the given repository will also be searched.
    """
    mods = []

    for file in _iterate_pybuilds(atom, repo_name):
        mod = safe_load_file(file)

        if mod is None:
            continue

        mods.append(mod)

    if repo_name is None and env.PREFIX_NAME:
        installed = load_installed_pkg(atom)
        # Ignore the name, in case it was moved
        if installed and atom_sat(installed.ATOM, atom, ignore_name=True):
            mods.append(installed)

    return mods


def load_all(
    *, repo_name: Optional[str] = None, only_repo_root: Optional[str] = None
) -> Generator[Pybuild, None, None]:
    """
    Loads all packages.

    args:
        repo_name: If specified, only loads packages accessible from this repository \
                   (including its masters)
        only_repo_root: If specified, only loads packages found within the given \
                        repository tree
    """
    for file in _iterate_pybuilds(repo_name=repo_name, only_repo_root=only_repo_root):
        mod = safe_load_file(file)
        if mod is None:
            continue

        yield mod


# TODO: Deprecated. Remove in 3.0
@prefix_aware_cache
def _load_local_pkg(package_path: str) -> InstalledPybuild:
    name = os.path.basename(package_path)
    # Use config to auto-detect special files such as plugins
    install_dir = InstallDir(".")
    flags = 0
    if get_config().get("CASE_INSENSITIVE_FILES"):
        flags = re.IGNORECASE

    def add_files(file_type, pattern, base_dir):
        if not pattern and os.path.exists(base_dir):
            getattr(install_dir, file_type).append(
                File(os.path.relpath(base_dir, package_path))
            )
        else:
            component, _, pattern = pattern.partition("/")
            for path in os.listdir(base_dir):
                if re.match(component, path, flags=flags):
                    add_files(file_type, pattern, os.path.join(base_dir, path))

    for file_type, pattern in get_config().get("LOCAL_FILES", {}).items():
        setattr(install_dir, file_type, [])
        add_files(file_type, pattern, package_path)

    return InstalledPybuild(
        FQAtom(f"local/{name}-0::installed"),
        INSTALL_DIRS=[install_dir],
        FILE=package_path,
        PROPERTIES="local",
        _PYBUILD_VER=1,
        TIER="a",
        DATA_OVERRIDES="",
        REPO="",
    )


def load_all_installed() -> Generator[InstalledPybuild, None, None]:
    """
    Returns a flat set of all installed packages
    """
    for path in _iterate_installed():
        mod = cast(Optional[InstalledPybuild], safe_load_file(path, installed=True))
        if mod:
            yield mod

    # TODO: Deprecated. Remove in 3.0
    local_dir = os.path.join(env.prefix().LOCAL_MODS)
    if os.path.exists(local_dir):
        for subdir in os.listdir(local_dir):
            path = os.path.join(env.prefix().LOCAL_MODS, subdir)
            if os.path.isdir(path):
                yield _load_local_pkg(path)


def load_all_installed_map() -> Dict[str, List[InstalledPybuild]]:
    """
    Returns every single installed mod in the form of a map from their simple mod name
    to their mod object
    """
    mods: Dict[str, List[InstalledPybuild]] = {}
    for mod in load_all_installed():
        if mods.get(mod.PN) is None:
            mods[mod.PN] = [mod]
        else:
            mods[mod.PN].append(mod)
    return mods


def load_file(path: str, installed: bool = False) -> Pybuild:
    """Loads the pybuild at the given path"""
    return __load_mod_from_dict_cache(path, installed=installed)


@__safe_load
def safe_load_file(path: str, installed: bool = False) -> Optional[Pybuild]:
    """
    Loads the pybuild at the given path

    :returns: The pybuild, or None if it could not be loaded
    """
    return load_file(path, installed)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Config sorting tests

Note that for the purposes of speed, mods are not removed between tests, and the
removal test occurs at the end. This means that the order of tests may matter,
and you should, when writing tests, assume an arbitrary configuration at the
beginning of the test, and attempt to place the test in such an order as to minimize
the number of changes required to get the desired configuration
"""

import os
from pathlib import Path

import pytest

from portmodlib.fs import _move2, ci_exists, match

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    setup_env("test")
    yield
    tear_down_env()


def test_ci_exists_nonexistant_file():
    """
    Tests that ci_exists does not recognize a file which does not exist
    """
    assert not ci_exists(os.path.abspath("File.txt"))


def test_ci_exists():
    """
    Tests that ci_exists recognizes files with case differences
    """
    open("File.TXT", "w").close()
    for pattern in ("File.TXT", "file.txt", "FILE.TXT"):
        assert ci_exists(pattern, prefix=os.getcwd())
        assert ci_exists(os.path.abspath(pattern))
        assert ci_exists(pattern)


def test_ci_exists_different_directory():
    """
    Tests that ci_exists recognizes files with case differences
    when the file is not relative to the current working directory
    """
    open("File.TXT", "w").close()
    parent = os.getcwd()
    os.makedirs("test")
    os.chdir("test")
    for pattern in ("File.TXT", "file.txt", "FILE.TXT"):
        assert ci_exists(pattern, prefix=parent)
        assert not ci_exists(pattern)

    os.chdir(parent)
    os.remove("File.TXT")


def test_match():
    """Tests that fs.match matches directories respecting path components"""
    assert match(Path("foo.txt"), "*.txt")
    assert not match(Path("bar/foo.txt"), "*.txt")
    assert match(Path("bar/baz/foo.txt"), "bar/*/*.txt")
    assert not match(Path("bar/baz/foo.txt"), "bar/*")


def test_match_recursive():
    """Tests that fs.match matches directories recursively"""
    assert match(Path("foo.txt"), "**/*.txt")
    assert match(Path("bar/foo.txt"), "**/*.txt")
    assert match(Path("bar/baz/foo.txt"), "**/*.txt")
    assert match(Path("bar/baz/foo.txt"), "**/*")
    assert match(Path("bar"), "**/*")
    assert not match(Path("bar/baz/foo.txt"), "**")
    assert not match(Path("bar"), "**")


def test_move2_symlink():
    os.symlink("foo", "bar")
    for file in os.scandir("."):
        if file.name == "bar":
            _move2(file, "baz")
    assert os.path.islink("baz")
    assert os.readlink("baz") == "foo"



# Copyright 2022 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the mod selection system
"""

import os

import pytest

from portmod.loader import load_pkg
from portmod.package import install_pkg
from portmod.vdb import VDB, vdb_path
from portmodlib.atom import Atom

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup_repo():
    yield setup_env("test")
    tear_down_env()


def test_vdb():
    """
    Tests that all files which need to be installed get added to the package database

    Tests that there are no untracked files in the VDB
    following package installation
    """
    pkg = load_pkg(Atom("test/test-install"))[0]
    install_pkg(pkg, {"foo", "bar"})

    root_path = os.path.join(vdb_path(), "test", "test-install")
    assert os.path.exists(
        os.path.join(root_path, "CONTENTS")
    ), "The CONTENTS file should be present"
    with open(os.path.join(root_path, "CONTENTS")) as file:
        contents = file.readlines()
        assert any(
            os.path.join("pkg", "test", "test-install", "Foo") in line
            for line in contents
        ), "The file installed should be registered in CONTENTS"

    assert os.path.exists(
        os.path.join(root_path, "files", "file")
    ), "The file included in the files directory should be installed"

    assert os.path.exists(
        os.path.join(root_path, "environment.xz")
    ), "The environment file should be installed"

    assert os.path.exists(
        os.path.join(root_path, "Manifest")
    ), "The Manifest file should be installed"

    assert os.path.exists(
        os.path.join(root_path, os.path.basename(pkg.FILE))
    ), "The pybuild file should be installed"

    assert os.path.exists(
        os.path.join(root_path, "REPO")
    ), "The REPO file should be installed"
    with open(os.path.join(root_path, "REPO")) as file:
        assert file.read().strip() == "test", 'The installed REPO should be "test"'

    assert os.path.exists(
        os.path.join(root_path, "USE")
    ), "The USE file should be installed"
    with open(os.path.join(root_path, "USE")) as file:
        assert set(file.read().strip().split()) == {
            "foo",
            "bar",
        }, "The USE flags should be correct"

    with VDB() as vdb:
        assert not vdb.git.ls_files(others=True, exclude_standard=True)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Interface for installing and removing packages"""

import argparse
import os
import sys
import traceback
from logging import error, info, warning
from shutil import move, rmtree

from portmod._deps import DepError, PackageDoesNotExist
from portmod.download import get_filename
from portmod.globals import env
from portmod.loader import AmbiguousAtom, load_all
from portmod.lock import exclusive_lock
from portmod.merge import configure, deselect, global_updates
from portmod.news import display_unread_message
from portmodlib.atom import InvalidAtom
from portmodlib.fs import onerror
from portmodlib.l10n import l10n

from . import atom_metavar


def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def filter_mods(mods):
    atoms = []
    os.makedirs(env.DOWNLOAD_DIR, exist_ok=True)

    for mod in mods:
        if os.path.isfile(mod):
            for atom in load_all():
                for source in atom.get_source_manifests(matchall=True):
                    if source.check_file(mod):
                        move(mod, get_filename(source.name))
                        atoms.append(atom.ATOM)
            if os.path.exists(mod):
                warning(l10n("no-package-for-file", file=os.path.basename(mod)))
        else:
            atoms.append(mod)

    return atoms


def add_merge_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "merge",
        help=l10n("merge-help"),
        description=l10n("merge-desc"),
        parents=parents,
        conflict_handler="resolve",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "packages",
        metavar=atom_metavar(archive=True, sets=True),
        help=l10n("package-help"),
        nargs="*",
    )
    parser.add_argument(
        "--ignore-default-opts",
        help=l10n("ignore-default-opts-help"),
        action="store_true",
    )
    parser.add_argument(
        "-c", "--depclean", help=l10n("depclean-help"), action="store_true"
    )
    parser.add_argument(
        "-x", "--auto-depclean", help=l10n("auto-depclean-help"), action="store_true"
    )
    parser.add_argument(
        "-C", "--unmerge", help=l10n("unmerge-help"), action="store_true"
    )
    parser.add_argument(
        "-1", "--oneshot", help=l10n("oneshot-help"), action="store_true"
    )
    parser.add_argument("-O", "--nodeps", help=l10n("nodeps-help"), action="store_true")
    parser.add_argument("-u", "--update", help=l10n("update-help"), action="store_true")
    parser.add_argument(
        "-n", "--noreplace", help=l10n("noreplace-help"), action="store_true"
    )
    parser.add_argument("-N", "--newuse", help=l10n("newuse-help"), action="store_true")
    parser.add_argument(
        "-e", "--emptytree", help=l10n("emptytree-help"), action="store_true"
    )
    parser.add_argument("-D", "--deep", help=l10n("deep-help"), action="store_true")
    parser.add_argument(
        "-w",
        "--select",
        type=str2bool,
        nargs="?",
        const=True,
        default=None,
        metavar=l10n("yes-or-no"),
        help=l10n("merge-select-help"),
    )
    parser.add_argument(
        "--deselect",
        type=str2bool,
        nargs="?",
        const=True,
        default=None,
        metavar=l10n("yes-or-no"),
        help=l10n("merge-deselect-help"),
    )
    parser.add_argument("--sort-vfs", help=argparse.SUPPRESS, action="store_true")
    parser.add_argument("--debug", help=l10n("merge-debug-help"), action="store_true")

    parser.set_defaults(func=merge_main)


@exclusive_lock()
def merge_main(args):
    atoms = filter_mods(args.packages)
    env.DEBUG = args.debug

    if args.nodeps and args.depclean:
        error(l10n("nodeps-depclean"))
        sys.exit(1)

    if atoms or args.depclean:
        # If deselect is supplied (is not None), only deselect if not removing.
        # If removing, remove normally, but deselect depending on supplied value.
        if args.deselect and not (args.unmerge or args.depclean):
            deselect(atoms)
        else:
            try:
                configure(
                    atoms,
                    delete=args.unmerge,
                    depclean=args.depclean,
                    oneshot=args.oneshot,
                    verbose=args.verbose,
                    update=args.update or args.newuse or args.noreplace,
                    nodeps=args.nodeps,
                    deselect=args.deselect,
                    select=args.select,
                    auto_depclean=args.auto_depclean,
                    deep=args.deep,
                    emptytree=args.emptytree,
                )

                # Note: When execeptions occur, TMP_DIR should be preserved
                if not env.DEBUG and os.path.exists(env.TMP_DIR):
                    rmtree(env.TMP_DIR, onerror=onerror)
                    info(">>> " + l10n("cleaned-up", dir=env.TMP_DIR))
            except (InvalidAtom, PackageDoesNotExist, AmbiguousAtom, DepError) as e:
                if args.debug:
                    traceback.print_exc()
                error(f"{e}")
                sys.exit(1)

    if args.sort_vfs:
        global_updates()

    display_unread_message()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import sys
from logging import error
from subprocess import CalledProcessError

from portmod.execute import sandbox_execute
from portmod.globals import env
from portmod.perms import Permissions
from portmodlib.l10n import l10n


def run(args):
    try:
        sandbox_execute(
            args.command,
            Permissions(
                global_read=True, rw_paths=[env.prefix().ROOT], tmp=env.TMP_DIR
            ),
        )
    except CalledProcessError as e:
        if args.debug:
            error(str(e))
        sys.exit(1)


def add_run_parser(subparsers, parents):
    parser = subparsers.add_parser("run", help=l10n("run-help"), parents=parents)
    parser.add_argument(
        "command",
        nargs="+",
        metavar=l10n("command-placeholder"),
        help=l10n("run-command-help"),
    )
    parser.set_defaults(func=run)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3
"""
Module that interacts with the various portmod config files

Files are stored both in the portmod local directory and in the profile directory tree,
with the user's config file overriding and extending defaults set by the profile
"""

import os
import sys
import warnings
from pathlib import Path
from typing import Any, Dict, Optional

from portmod.config.profiles import profile_exists, profile_parents
from portmod.functools import prefix_aware_cache
from portmod.globals import env
from portmod.win32 import get_personal

from .pyconf import (
    __COLLAPSE_KEYS,
    __OVERRIDE_KEYS,
    _create_config_placeholder,
    read_config,
)


# Note: this method should not be cached as otherwise it may be possible for a package
# to replace var with a symlink and then cause files to be written outside of the prefix
def variable_data_dir() -> str:
    """
    The directory in whch variable data is stored.

    I.e. all portmod information not installed by packages
    This is controlled with the VARIABLE_DATA profile variable and
    must be a path within the prefix ROOT.
    """
    # Resolve in case it includes symlinks outside of the prefix
    path = Path(env.prefix().ROOT, get_config_value("VARIABLE_DATA")).resolve()
    # is_relative_to is python3.9+
    # relative_to will raise a ValueError if the path is not relative to ROOT
    path.relative_to(env.prefix().ROOT)
    return str(path)


def get_config_value(key: str, default: Optional[Any] = None):
    """
    Parses the user's configuration, overriding defaults from their profile

    returns:
        The config value matching key, or default if it was not specified
    """
    return get_config().get(key, default)


@prefix_aware_cache
def get_config() -> Dict[str, Any]:
    """
    Parses the user's configuration, overriding defaults from their profile
    """
    total_config: Dict[str, Any] = {
        # Default cannot be set in profile due to the value depending on platform
        "PLATFORM": sys.platform,
    }

    for attr in os.environ:
        total_config[attr] = os.environ[attr]

    if env.PREFIX_NAME is not None:
        total_config["ARCH"] = env.prefix().ARCH
        total_config["ROOT"] = env.prefix().ROOT
        total_config["VARIABLE_DATA"] = "var"
        total_config["PROFILE_ONLY_VARIABLES"] = ["VARIABLE_DATA"]

    if sys.platform == "win32":
        total_config["PERSONAL"] = get_personal()

    if profile_exists():
        for parent in profile_parents():
            path = os.path.join(parent, "defaults.conf")
            if os.path.exists(path):
                total_config = read_config(path, total_config)

    if os.path.exists(env.GLOBAL_PORTMOD_CONFIG):
        total_config = read_config(env.GLOBAL_PORTMOD_CONFIG, total_config, user=True)

    if env.PREFIX_NAME:
        if os.path.exists(env.prefix().CONFIG):
            total_config = read_config(env.prefix().CONFIG, total_config, user=True)
        else:
            _create_config_placeholder()

    # Set defaults and add to os.environ (for use in portmodlib)
    for key in __OVERRIDE_KEYS:
        if not total_config.get(key):
            total_config[key] = ""
        elif isinstance(total_config[key], (str, bool)) and total_config[key]:
            os.environ[key] = str(total_config[key])

    for key in __COLLAPSE_KEYS:
        if not total_config.get(key):
            total_config[key] = set()
        elif isinstance(total_config[key], (str, bool)) and total_config[key]:
            os.environ[key] = str(total_config[key])

    # Note: VARIABLE_DATA needs to be set as an environment variable
    # so that we can access it both in the sandbox and outside with
    # a common interface
    for key in ["TMPDIR", "TEMP", "TMP", "VARIABLE_DATA"]:
        if total_config.get(key):
            os.environ[key] = total_config[key]

    return total_config


def config_to_string(config: Dict) -> str:
    """Prints the given dictionary config as a string"""
    lines = []
    for key in sorted(config):
        if isinstance(config[key], (list, set)):
            lines.append("{} = {}".format(key, " ".join(sorted(config[key]))))
        else:
            lines.append("{} = {}".format(key, config[key]))
    return "\n".join(lines)


def set_config_value(key: str, value: str, path: Optional[str] = None) -> Optional[str]:
    """
    Sets the given key-value pair in portmod.conf

    The previous value is returned, if any
    """
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        from redbaron import RedBaron
        from redbaron.nodes import AssignmentNode, NameNode

    if not path:
        path = env.prefix().CONFIG

    old = None
    string = '"' + value + '"'
    new_line = f"{key} = {string}"
    if os.path.exists(path):
        with open(path, "r") as file:
            node = RedBaron(file.read())

            for elem in node:
                if (
                    isinstance(elem, AssignmentNode)
                    and isinstance(elem.target, NameNode)
                    and elem.target.value == key
                ):
                    old = elem.value
                    elem.value = string

            if not old and node:
                node.append(new_line)
            elif not node:
                node = RedBaron(new_line)

            with open(path, "w") as file:
                file.write(node.dumps())
    else:
        with open(path, "w") as file:
            file.write(new_line)

    get_config.cache_clear()
    return old



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests modifying user use flags
"""

import sys
from io import StringIO

import pytest

import portmodlib.log as log
from portmod._cli.main import main
from portmod.globals import env
from portmod.repos import get_repos
from portmod.sync import sync

from .env import setup_env, tear_down_env, unset_profile


@pytest.fixture(scope="module", autouse=True)
def setup():
    """sets up and tears down test environment"""
    dictionary = setup_env("test")
    env.REPOS = get_repos()
    # Unset profile so news is marked as old
    unset_profile()
    yield dictionary
    tear_down_env()


def test_logging():
    """Tests that verbose and quiet control output verbosity"""
    oldargs = sys.argv
    sync(env.REPOS)
    sys.argv = ["portmod", "sync", "--quiet"]
    output = StringIO()

    # Note: Logging cannot be redirected with redirect_stderr,
    # as the output stream is set on logger initialization
    log._LOG_HANDLER.__init__(output)  # type: ignore

    main()

    assert not output.getvalue()

    sys.argv = ["portmod", "sync", "--verbose"]
    output = StringIO()
    log._LOG_HANDLER.__init__(output)  # type: ignore
    main()

    assert output.getvalue()

    sys.argv = oldargs
    log._LOG_HANDLER.__init__(sys.stderr)  # type: ignore



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Interface for interacting with installed modules
"""

import glob
import os
import shutil
from logging import info
from types import SimpleNamespace
from typing import Generator, List, Optional

from portmodlib.fs import onerror
from portmodlib.l10n import l10n

from .cfg_protect import get_redirections, handle_cfg_protect
from .config import get_config_value, variable_data_dir
from .globals import env
from .loader import load_module


class ModuleState(SimpleNamespace):
    TEMP: str
    ROOT: str
    CACHE: str


def do_func(state, func, args=None):
    if args is None:
        func(state)
    else:
        func(state, args)


class ModuleFunction:
    """Function defined by a module"""

    name: str

    def __init__(
        self,
        name: str,
        desc: Optional[str],
        do,
        options,
        parameters,
        state: ModuleState,
    ):
        self.name = name
        self.desc = desc
        self.__do__ = do
        self.state = state
        if options is not None:
            self.options = options
        else:
            self.options = []
        if parameters is not None:
            self.parameters = parameters
        else:
            self.parameters = []

    def do(self, args):
        """Execute action"""
        do_func(
            self.state,
            self.__do__,
            {key: getattr(args, key) for key in self.options},
        )
        handle_cfg_protect(get_redirections())

    def do_noargs(self):
        """Execute action without arguments"""
        do_func(self.state, self.__do__)


class Module:
    """Base module object"""

    def __init__(self, name: str, desc: str, funcs: List[ModuleFunction], state):
        self.funcs = {func.name: func for func in funcs}
        self.name = name
        self.desc = desc
        self.state = state
        os.makedirs(state.TEMP, exist_ok=True)
        os.makedirs(state.CACHE, exist_ok=True)

    def update(self):
        if "update" in self.funcs:
            self.funcs["update"].do_noargs()

    def add_parser(self, parsers, parents):
        parser = parsers.add_parser(self.name, help=self.desc, parents=parents)
        this_subparsers = parser.add_subparsers()
        for func in self.funcs.values():
            if func.name == "update":
                continue
            func_parser = this_subparsers.add_parser(func.name, help=func.desc)
            for option, parameter in zip(func.options, func.parameters):
                func_parser.add_argument(option, help=parameter)
            func_parser.set_defaults(func=func.do)

        def help_func(args):
            parser.print_help()

        parser.set_defaults(func=help_func)
        self.arg_parser = parser

        return self.arg_parser

    def prerm(self):
        if "prerm" in self.funcs:
            self.funcs["prerm"].do_noargs()

    def cleanup(self):
        shutil.rmtree(self.state.TEMP, onerror=onerror)


def get_state(module_name: str) -> ModuleState:
    return ModuleState(
        TEMP=os.path.join(env.TMP_DIR, "modules", module_name),
        ROOT=os.path.join(env.prefix().ROOT),
        CACHE=os.path.join(env.prefix().CACHE_DIR, "modules", module_name),
    )


def iterate_modules() -> Generator[Module, None, None]:
    """Returns a generator which produces all modules"""
    modules_dir = get_config_value("MODULEPATH")
    if modules_dir:
        modules_path = os.path.join(env.prefix().ROOT, modules_dir)
        for module_file in glob.glob(os.path.join(modules_path, "*.pmodule")):
            module_name, _ = os.path.splitext(os.path.basename(module_file))
            module = load_module(module_file, get_state(module_name))
            yield module
            module.cleanup()


def module_prerm(path: str):
    module_name, _ = os.path.splitext(os.path.basename(path))
    module = load_module(path, get_state(module_name))
    module.prerm()
    module.cleanup()


def update_modules():
    """Runs update function (if present) on all installed modules"""
    for module in iterate_modules():
        info(l10n("updating-module", name=module.name))
        module.update()

    handle_cfg_protect(get_redirections())


def add_parsers(parsers, parents) -> List[Module]:
    """Adds parsers for the modules to the given argument parser"""
    modules = []
    for module in iterate_modules():
        module.add_parser(parsers, parents)
        modules.append(module)
    return modules


def require_module_updates():
    """
    Creates a file that indicates that modules need to be updated
    """
    open(os.path.join(variable_data_dir(), ".modules_need_updating"), "a").close()


def clear_module_updates():
    """Clears the file indicating that modules need updating"""
    path = os.path.join(variable_data_dir(), ".modules_need_updating")
    if os.path.exists(path):
        os.remove(path)


def modules_need_updating():
    """Returns true if changes have been made since the config was sorted"""
    return os.path.exists(os.path.join(variable_data_dir(), ".modules_need_updating"))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import sys

import pytest

from portmod._cli.main import main
from portmod.globals import env
from portmod.merge import configure
from portmod.modules import get_redirections, iterate_modules

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    os.makedirs(env.prefix().CONFIG_PROTECT_DIR, exist_ok=True)
    yield dictionary
    tear_down_env()


def test_module(setup):
    """Tests that modules work as expected"""
    configure(["test/test-module"], oneshot=True)
    assert os.path.exists(
        os.path.join(env.prefix().CONFIG_PROTECT_DIR, "foo.cfg_protect")
    )
    assert get_redirections()


def test_module_params(setup):
    configure(["test/test-module"], oneshot=True)
    for module in iterate_modules():
        module_function = module.funcs.get("add")
        assert module_function
        assert module_function.name == "add"
        assert module_function.desc == "Add to list"
        assert module_function.options == ["item"]
        assert module_function.parameters == ["item to add to the list"]


def test_module_cli(setup):
    """Tests that the CLI module interface works"""
    configure(["test/test-module"], oneshot=True)
    sys.argv = ["portmod", "test", "select", "test-module", "list"]
    main()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import enum
import os
from typing import Any

from .globals import download_dir


class EnumMeta(enum.EnumMeta):
    def __contains__(cls, item: Any):
        return item in {member.value for member in cls.__members__.values()}  # type: ignore


class HashAlg(enum.Enum, metaclass=EnumMeta):
    """
    Class for interacting with supported hash algorithms that can be used in manifests
    """

    BLAKE2B = "BLAKE2B"
    MD5 = "MD5"
    SHA512 = "SHA512"
    SHA256 = "SHA256"
    BLAKE3 = "BLAKE3"

    def __lt__(self, other):
        return self.value < other.value


def get_archive_basename(archive: str) -> str:
    """Returns archive name minus extension(s)"""
    basename, _ = os.path.splitext(archive)
    # Hacky way to handle tar.etc having multiple extensions
    if basename.endswith("tar"):
        basename, _ = os.path.splitext(basename)
    return basename


class Source:
    """
    Class used for storing information about download files without manifest information
    """

    def __init__(self, url: str, name: str):
        self.url = url
        self.name = name
        self.path = os.path.join(download_dir(), name)
        self.basename = get_archive_basename(name)

    def __repr__(self):
        return self.url

    def to_json(self):
        return {"url": self.url, "name": self.name}

    def __eq__(self, other):
        if not isinstance(other, Source):
            return False
        return self.url == other.url and self.name == other.name

    def __hash__(self):
        return hash((self.url, self.name))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import logging
import os
import shutil
import sys
import urllib.parse
from base64 import b16encode, b64decode
from logging import info
from typing import AbstractSet, Dict, Iterable, List, Optional, Set

from portmod.config import get_config
from portmod.config.license import _get_pkg_license_groups
from portmod.globals import env
from portmod.parsers.manifest import FileType
from portmod.pybuild import Pybuild
from portmod.source import LocalHashError, SourceManifest
from portmodlib.fs import get_hash
from portmodlib.l10n import l10n
from portmodlib.portmod import directories
from portmodlib.source import HashAlg, Source
from portmodlib.usestr import use_reduce


class RemoteHashError(Exception):
    """Exception indicating an unexpected download file"""


class SourceUnfetchable(Exception):
    """Exception indicating an unexpected download file"""


def get_filename(basename: str, path: Optional[str] = None) -> str:
    """
    By default returns the location of the local cached version of the source
    file corresponding to the given name. The file may or may not exist.
    @return full path of file combined from path and basename
    """
    if not path:
        path = env.DOWNLOAD_DIR
    return os.path.join(path, basename)


def mirrorable(mod: Pybuild, enabled_use: Optional[AbstractSet[str]] = None) -> bool:
    """
    Returns whether or not a mod can be mirrored
    """
    # FIXME: Determine which sources have use requirements
    # that don't enable mirror or fetch
    if "mirror" not in mod.RESTRICT and "fetch" not in mod.RESTRICT:
        redis = _get_pkg_license_groups(mod).get("REDISTRIBUTABLE", [])

        def is_license_redis(group):
            if not group:
                # No license is considered equivalent to all-rights-reserved
                return False
            if isinstance(group, str):
                return group in redis
            if group[0] == "||":
                return any(is_license_redis(li) for li in group)

            return all(is_license_redis(li) for li in group)

        if enabled_use and is_license_redis(
            use_reduce(mod.LICENSE, enabled_use, opconvert=True)
        ):
            return True
        if not enabled_use and is_license_redis(
            use_reduce(mod.LICENSE, opconvert=True, matchall=True)
        ):
            return True

    return False


def fetchable(
    mod: Pybuild, enabled_use: AbstractSet[str] = frozenset(), matchall: bool = False
) -> List[SourceManifest]:
    """
    Returns the list of fetchable sources associated with a mod
    A source can be fetched if it is mirrorable,
    or if it has a URI
    """
    restricted = use_reduce(mod.RESTRICT, enabled_use, flat=True, matchall=matchall)
    if "fetch" in restricted:
        return []

    if "mirror" not in restricted and mirrorable(mod, enabled_use):
        return mod.get_source_manifests(enabled_use)

    can_fetch = []
    for source in mod.get_source_manifests(enabled_use):
        parsedurl = urllib.parse.urlparse(source.url)
        if parsedurl.scheme:
            can_fetch.append(source)

    return can_fetch


def progress_write(request, filepath, start=0, mode="wb", size: Optional[int] = None):
    if sys.stdout.isatty():
        try:
            from progressbar import GranularBar as Bar
        except ImportError:
            from progressbar import Bar
        from progressbar import (
            ETA,
            DataSize,
            FileTransferSpeed,
            Percentage,
            ProgressBar,
        )

        bytes = start
        bar = ProgressBar(
            widgets=[
                Percentage(),
                " ",
                Bar(),
                " ",
                ETA(),
                " ",
                FileTransferSpeed(),
                " ",
                DataSize(),
            ],
            redirect_stdout=True,
            max_value=size,
        )

        bar.start()
        with open(filepath, mode=mode) as file:
            for buf in request.iter_content(1024):
                if buf:
                    file.write(buf)
                    bytes += len(buf)
                    bar.update(bytes)
        bar.finish()
    else:
        with open(filepath, mode=mode) as file:
            for buf in request.iter_content(1024):
                if buf:
                    file.write(buf)


def get_remote_size(request):
    if "Content-Length" in request.headers:
        return int(request.headers["Content-Length"].strip())
    else:
        return None


def get_remote_hashes(request):
    header = "X-Goog-Hash"
    remote_hashes = {}

    if header in request.headers:
        for remote_hash in request.headers[header].split(", "):
            hash = remote_hash.split("=", maxsplit=1)
            remote_hashes[hash[0]] = hash[1]
        if "md5" in remote_hashes:
            remote_hashes["md5"] = (
                b16encode(b64decode(remote_hashes["md5"])).decode("utf-8").lower()
            )

    return remote_hashes


def download(
    url: str,
    dest_name: str,
    *,
    size: Optional[int] = None,
    hashes: Dict[HashAlg, str] = {},
    dest_path: Optional[str] = None,
    check_only: bool = False,
):
    """
    Downloads the given url to the dest_path
    """
    # Slow imports
    import requests

    print(l10n("fetching", url=url))

    if not dest_path:
        dest_path = env.DOWNLOAD_DIR

    os.makedirs(dest_path, exist_ok=True)

    req = requests.get(url, stream=True)
    req.raise_for_status()

    precheck = False
    remote_size = get_remote_size(req)
    remote_hashes = get_remote_hashes(req)

    if "md5" in remote_hashes and HashAlg.MD5 in hashes:
        precheck = True
        remote_hash = remote_hashes["md5"]
        if hashes[HashAlg.MD5] != remote_hash:
            raise RemoteHashError(
                l10n(
                    "remote-hash-mismatch", hash1=hashes[HashAlg.MD5], hash2=remote_hash
                )
            )

    if remote_size is not None and size is not None and remote_size != size:
        raise RemoteHashError(
            l10n("remote-size-mismatch", our_size=size, remote_size=remote_size)
        )

    if not check_only:
        file = get_filename(dest_name, dest_path)
        progress_write(req, file, size=remote_size or size)

        if remote_size:
            filesize = os.path.getsize(file)
            while filesize < remote_size:
                req = requests.get(
                    url, stream=True, headers={"Range": "bytes=%d-" % filesize}
                )
                progress_write(req, file, start=filesize, mode="ab", size=remote_size)
                filesize = os.path.getsize(file)

        # Skip post check if we did a pre-check using a local hash
        if not precheck and "md5" in remote_hashes:
            localhash = get_hash(get_filename(dest_name), (HashAlg.MD5,))[0]
            if localhash != remote_hashes["md5"]:
                # Try again
                download(url, dest_name, size=size, hashes=hashes, dest_path=dest_path)


def get_download(source: Source) -> Optional[str]:
    src = find_download(source)
    if src:
        dest = get_filename(source.name)
        if src == dest:
            return dest

        info(l10n("file-moving", src=src, dest=dest))
        shutil.move(src, dest)
        return dest
    return None


def find_download(source: Source) -> Optional[str]:
    """
    Determines if the given file is in the cache.
    The file must match both name and checksum
    @return path to donloaded file
    """

    def find_by_name(directory: str) -> Optional[str]:
        if os.path.exists(directory):
            src = os.path.join(directory, source.name)
            if os.path.exists(src) and (
                not isinstance(source, SourceManifest) or source.check_file(src)
            ):
                return src

            # If a file with spaces otherwise matches use it
            for file in os.listdir(directory):
                src = os.path.join(directory, file)
                if file.replace(" ", "_") == source.name:
                    try:
                        if not isinstance(source, SourceManifest) or source.check_file(
                            src, raise_ex=True
                        ):
                            return src
                    except LocalHashError as e:
                        logging.warning(
                            l10n(
                                "possible-local-hash-mismatch",
                                filename=file,
                                name=source.name,
                            )
                            + f": {e}"
                        )
        return None

    return (
        find_by_name(env.DOWNLOAD_DIR)
        or find_by_name(
            directories.download_dir
            or os.path.expanduser(os.path.join("~", "Downloads"))
        )
        or find_by_name(os.environ.get("DOWNLOADS", ""))
    )


def is_downloaded(mod: Pybuild, enabled_use: Set[str]) -> bool:
    """
    Returns true if all the mod's sources can be found
    """
    for source in mod.get_source_manifests(enabled_use):
        cached = find_download(source)
        if cached is None:
            return False
    return True


def download_and_check(url: str, source: SourceManifest, *, try_again: bool = True):
    download(url, source.name, size=source.size, hashes=source.hashes)
    filename = get_filename(source.name)

    try:
        source.check_file(filename, raise_ex=True)
    except LocalHashError as e:
        if try_again:  # Try again once if download fails
            get_hash.cache_clear()
            logging.error(e)
            print(l10n("retrying-download", url=url))
            download_and_check(url, source, try_again=False)
        else:
            raise


def download_source(mod, source: Source, *, check_remote: bool = False) -> str:
    """
    Downloads the given source file.
    @return the path to the downloaded source file
    """
    # Slow import
    import requests

    cached = get_download(source)
    fetch = "fetch" not in mod.get_restrict()
    mirror = (
        "mirror" not in mod.get_restrict()
        and "fetch" not in mod.get_restrict()
        and mirrorable(mod)
    )

    if cached:
        # Download is in cache. Nothing to do.
        return cached
    elif not fetch:
        # Mod cannot be fetched and is not already in cache. abort.
        raise SourceUnfetchable(l10n("source-unfetchable", source=source))
    else:
        parsedurl = urllib.parse.urlparse(source.url)

        # Download archive
        filename = get_filename(source.name)

        def _download_source_inner(url: str):
            if isinstance(source, SourceManifest):
                if check_remote:
                    download(
                        url,
                        source.name,
                        size=source.size,
                        hashes=source.hashes,
                        check_only=True,
                    )
                else:
                    download_and_check(url, source)
            else:
                download(url, source.name, check_only=check_remote)

        if mirror:
            PORTMOD_MIRRORS = get_config()["PORTMOD_MIRRORS"].split()
            for mirror_url in PORTMOD_MIRRORS:
                try:
                    url = urllib.parse.urljoin(mirror_url, source.name)

                    _download_source_inner(url)

                    return filename
                except requests.exceptions.HTTPError as err:
                    if parsedurl.scheme:
                        logging.info(f"{err}")
                    else:  # If there is no original source, warn instead of info
                        logging.warning(f"{err}")

        if parsedurl.scheme != "":
            _download_source_inner(source.url)

            return filename

        raise SourceUnfetchable(l10n("source-unfetchable", source=source))


def download_mod(
    mod: Pybuild, enabled_flags: AbstractSet[str], matchall=False
) -> List[SourceManifest]:
    """
    Downloads missing sources for the given mod in its current USE configuration
    @return A list of paths of the sources for the mod
    """
    download_list = []
    if matchall:
        sources = mod.get_source_manifests(matchall=True)
    else:
        sources = mod.get_source_manifests(enabled_flags)

    for source in sources:
        download_source(mod, source)

        download_list.append(source)

    return download_list


def get_total_download_size(mods: Iterable[Pybuild]):
    download_bytes = 0
    for mod in mods:
        manifest_file = mod.get_manifest()
        for manifest in manifest_file.entries.values():
            if manifest.file_type == FileType.DIST:
                download_bytes += manifest.size

    return "{:.3f} MiB".format(download_bytes / 1024 / 1024)


def get_download_size(packages: Dict[Pybuild, Set[str]]) -> float:
    """
    Returns the download size for the packages in the given configuration

    args:
        packages: A dictionary mapping each package to check to its enabled USE flags

    returns:
        The size of the downloads in MiB
    """
    download_bytes = 0
    for mod, enabled_use in packages.items():
        sources = mod.get_source_manifests(enabled_use)
        for source in sources:
            if find_download(source) is None:
                download_bytes += source.size

    return download_bytes / 1024 / 1024



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import re
import sys
from logging import info, warning
from typing import (
    Callable,
    Dict,
    Generator,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    cast,
)

from portmod.functools import install_cache, prefix_aware_cache
from portmod.globals import env
from portmod.parsers.flags import add_flag, collapse_flags, get_flags, remove_flag
from portmod.pybuild import Pybuild
from portmod.repo import get_repo
from portmod.repo.metadata import get_use_flag_atom_aliases
from portmodlib.atom import Atom, QualifiedAtom
from portmodlib.l10n import l10n
from portmodlib.usestr import use_reduce

from . import get_config, read_config, set_config_value
from .profiles import profile_parents
from .textures import select_texture_size


class InvalidFlag(Exception):
    """Exception indicating an invalid use flag"""

    def __init__(self, flag: str, atom: Optional[Atom] = None):
        if atom:
            super().__init__(l10n("invalid-flag-atom", flag=flag, atom=atom))
        else:
            super().__init__(l10n("invalid-flag", flag=flag))


@prefix_aware_cache
def _get_use_defaults(
    pkg: Pybuild, is_installed: Optional[Callable[[QualifiedAtom], bool]] = None
) -> Set:
    """
    Returns a set of USE flags, each prepended with a plus, minus, or nothing,
    """

    from portmod.loader import load_installed_pkg

    if is_installed is None:

        def is_installed(atom):
            pkg = load_installed_pkg(atom)
            if pkg is not None:
                if atom.USE <= pkg.INSTALLED_USE:
                    return True
            return False

    assert is_installed is not None

    defaults: Set[str] = set()

    # Default values in IUSE have the lowest priority
    for x in pkg.IUSE:
        if x.startswith("+"):
            defaults.add(x[1:])
        else:
            defaults.add("-" + x)

    # Aliases override the default value provided in IUSE,
    # Unless they are prototypical aliases
    #   E.g. pkg "cat/foo" where "bar" is aliased to "cat/foo[bar]"
    aliases: Mapping[str, QualifiedAtom] = {}
    if "local" not in pkg.PROPERTIES:
        aliases = get_use_flag_atom_aliases(get_repo(pkg.REPO).location)
    aliased_flags = set()
    for flag in pkg.IUSE_EFFECTIVE:
        if (
            flag in aliases
            and is_installed(aliases[flag])
            and aliases[flag].CPN != pkg.CPN
        ):
            aliased_flags.add(flag)
    return collapse_flags(defaults, aliased_flags)


@prefix_aware_cache
def get_use_defaults(pkg: Pybuild) -> Dict[str, bool]:
    """
    Returns a dictionary containing the default states of the
    given Pybuild's USE flags.
    """

    default_flags = _get_use_defaults(pkg)
    defaults_dict: Dict[str, bool] = {}

    # Default values in IUSE have the lowest priority
    for flag in default_flags:
        if flag.startswith("+"):
            defaults_dict[flag[1:]] = True
        elif flag.startswith("-"):
            defaults_dict[flag[1:]] = False
        else:
            defaults_dict[flag] = False

    return defaults_dict


@install_cache
def get_use(
    pkg: Pybuild, is_installed: Optional[Callable[[QualifiedAtom], bool]] = None
) -> Tuple[Set[str], Set[str]]:
    """
    Returns a list of enabled and a list of disabled use flags

    The disabled flag list only contains the flags which have been explicitly disabled.
    Any flags missing from both lists should be considered implicitly disabled.
    """

    defaults = _get_use_defaults(pkg, is_installed)

    GLOBAL_USE = get_config().get("USE", [])
    use: Set[str] = get_use_expand_flags()
    use = collapse_flags(use, GLOBAL_USE)
    for parent_dir in profile_parents():
        for use_file in _iter_files(parent_dir, "package", "use"):
            flags = get_flags(use_file, pkg.ATOM)
            use = collapse_flags(use, flags)

    # User config is the last added, overriding profile flags
    use = collapse_flags(use, get_user_use(pkg.ATOM))

    # Finally, apply environment variables (technically, we do this twice,
    # as GLOBAL_USE also includes these, but we need to apply them again to
    # ensure they override
    use = collapse_flags(use, os.environ.get("USE", []))

    # Forced use flags must be collapsed last to ensure that the
    # forced flags override any other values
    use = collapse_flags(use, get_forced_use(pkg.ATOM))

    enabled_use = {x for x in use if not x.startswith("-")}
    disabled_use = {x.lstrip("-") for x in use if x.startswith("-")}

    enabled_use |= {
        x for x in defaults if not x.startswith("-") and x not in disabled_use
    }

    enabled_use = enabled_use.intersection(pkg.IUSE_EFFECTIVE)
    disabled_use = disabled_use.intersection(pkg.IUSE_EFFECTIVE)

    texture_sizes = use_reduce(
        pkg.TEXTURE_SIZES, enabled_use, disabled_use, token_class=int
    )
    texture_size = select_texture_size(texture_sizes)
    if texture_size is not None:
        found = None
        for useflag in enabled_use:
            if useflag.startswith("texture_size_"):
                if not found:
                    found = useflag
                elif useflag != found:
                    raise Exception(
                        l10n(
                            "multiple-texture-flags",
                            flag1=use,
                            flag2=found,
                            atom=pkg.ATOM,
                        )
                    )
        enabled_use.add(found or "texture_size_{}".format(texture_size))

    return (enabled_use, disabled_use)


def get_use_expand_flags() -> Set[str]:
    """Returns all currently enabled USE_EXPAND flags"""
    flags = set()
    for use in get_config().get("USE_EXPAND", []):
        if not use.startswith("-"):
            for flag in get_config().get(use, "").split():
                flags.add(f"{use.lower()}_{flag}")
    return flags


def get_use_expand(pkg: Pybuild, use_expand: str) -> Tuple[Set[str], Set[str]]:
    """
    Returns the set of enabled flags for the given package and USE_EXPAND category
    """
    flags = {
        re.sub(f"^{use_expand.lower()}_", "", flag)
        for flag in pkg.IUSE_EFFECTIVE
        if flag.startswith(f"{use_expand.lower()}_")
    }
    enabled_use = pkg.get_use()
    enabled_expand = {
        re.sub(f"^{use_expand}_", "", val, flags=re.IGNORECASE)
        for val in enabled_use
        if val.startswith(use_expand.lower() + "_")
    }

    return enabled_expand, flags - enabled_expand


@prefix_aware_cache
def get_user_global_use() -> Set[str]:
    """Returns user-specified global use flags"""
    global_use = cast(
        Set[str], read_config(env.prefix().CONFIG, {}, user=True).get("USE", set())
    )

    # Also return global user-set USE_EXPAND values
    for use in get_config().get("USE_EXPAND", []):
        values = set(
            read_config(env.prefix().CONFIG, {}, user=True).get(use, "").split()
        )
        flags = {use.lower() + "_" + value for value in values}
        global_use |= flags

    return global_use


def _user_package_use_path() -> str:
    """Returns the path of the user's package.use"""
    return os.path.join(env.prefix().CONFIG_DIR, "package.use")


@prefix_aware_cache
def get_user_use(atom: Atom) -> Set[str]:
    """Returns user-specified use flags for a given mod"""
    use_file = _user_package_use_path()
    if os.path.exists(use_file):
        return get_flags(use_file, atom)

    return set()


def _iter_files(
    directory: str, basename: Optional[str], suffix: str
) -> Generator[str, None, None]:
    if basename:
        paths = [
            ".".join([basename, suffix]),
            ".".join([basename, sys.platform, suffix]),
        ]
    else:
        paths = [".".join([suffix]), ".".join([sys.platform, suffix])]

    for path in paths:
        path = os.path.join(directory, path)
        if os.path.exists(path):
            yield path


@prefix_aware_cache
def get_forced_use(atom: Atom) -> Set[str]:
    """Returns a list of forced use flags for the given mod"""
    force: Set[str] = set()
    for parent_dir in profile_parents():
        for force_file in _iter_files(parent_dir, None, "use.force"):
            flags = get_flags(force_file)
            force = collapse_flags(force, flags)

        for pkg_force_file in _iter_files(parent_dir, "package", "use.force"):
            flags = get_flags(pkg_force_file, atom)
            force = collapse_flags(force, flags)

    return force


def verify_use(flag: str, atom: Optional[Atom] = None) -> bool:
    """
    Verifies that the given flag is valid

    args:
        flag: Flag to be added
        atom: If present, checks for local use flags relevant to this atom.
              Otherwise, global use flags will be considered
    returns:
        Whether or not the use flag is valid given either the atom passed,
        or the global context
    """
    from portmod.repo.metadata import get_global_use

    from ..loader import load_pkg

    if atom:
        return any(flag in mod.IUSE_EFFECTIVE for mod in load_pkg(atom))
    else:
        return flag in [
            flag
            for repo in env.prefix().REPOS
            for flag in get_global_use(repo.location)
        ]


def add_use(
    flag: str,
    atom: Atom,
    disable: bool = False,
    comment: Sequence[str] = (),
    protect_file: Optional[str] = None,
):
    """
    Adds the given use flag to the user's configuration

    Args:
        flag: Flag to be added
        atom: Package the flag applies to
        disable: If True, the flag is added in its disabled form (prefixed with "-")
        comment: An optional comment to include with the flag override. Only meaningful
                 with an ``atom`` argument.
        protect: Whether or not the change should be protected by redirecting to a new
                 file and registering with the cfg_protect system to be updated when
                 cfg-update is run.
                 Only meaningful with an ``atom`` argument.
    """
    from portmod.loader import load_pkg

    disableflag = "-" + flag

    if not verify_use(flag, atom):
        raise InvalidFlag(flag, atom)

    if not any(flag in pkg.IUSE_EFFECTIVE for pkg in load_pkg(atom)):
        warning(l10n("invalid-use-flag-warning", flag=flag, atom=atom))

    use_file = protect_file or _user_package_use_path()

    if comment:
        with open(use_file, "a+", encoding="utf-8") as file:
            for line in comment:
                print(line, file=file)
            if disable:
                print(atom, disableflag, file=file)
            else:
                print(atom, flag, file=file)
        info(l10n("flag-add", flag=flag, atom=atom, file=use_file))
    else:
        if disable:
            remove_flag(use_file, atom, flag)
            add_flag(use_file, atom, disableflag)
        else:
            remove_flag(use_file, atom, disableflag)
            add_flag(use_file, atom, flag)
    get_user_use.cache_clear()
    get_use.cache_clear()


def add_global_use(flag: str, disable: bool = False):
    """
    Adds the given use flag to the user's global USE configuration

    Args:
        flag: Flag to be added
        disable: If True, the flag is added in its disabled form (prefixed with "-")
    """
    disableflag = "-" + flag

    if (disable and flag in get_config().get("USE", [])) or (
        not disable and disableflag in get_config().get("USE", [])
    ):
        remove_use(flag)

    if not verify_use(flag):
        raise InvalidFlag(flag)

    global_use = get_config()["USE"]

    if (not disable and flag not in global_use) or (
        disable and disableflag not in global_use
    ):
        if disable:
            info(l10n("adding-use-flag", flag="-" + flag))
            global_use.add(disableflag)
        else:
            info(l10n("adding-use-flag", flag=flag))
            global_use.add(flag)

        set_config_value("USE", " ".join(sorted(global_use)))
        get_user_global_use.cache_clear()
    else:
        if disable:
            warning(l10n("global-use-flag-already-disabled", flag=flag))
        else:
            warning(l10n("global-use-flag-already-enabled", flag=flag))
    get_use.cache_clear()


def remove_use(flag: str, atom: Optional[Atom] = None):
    """
    Removes the given use flag from the user's configuration

    This will remove both enabled and disabled flags

    Args:
        flag: Flag to be removed
        atom: If present, the flag will only be removed from configuration specifically
              affecting this package atom
    """

    disableflag = "-" + flag

    if not verify_use(flag, atom):
        warning(l10n("invalid-flag-atom", flag=flag, atom=atom))

    if atom is not None:
        use_file = _user_package_use_path()
        remove_flag(use_file, atom, flag)
        remove_flag(use_file, atom, disableflag)
        get_user_use.cache_clear()
    else:
        GLOBAL_USE = get_config()["USE"]

        if flag in GLOBAL_USE or disableflag in GLOBAL_USE:
            if flag in GLOBAL_USE:
                info(l10n("removing-use-flag", flag=flag))
                GLOBAL_USE.remove(flag)

            if disableflag in GLOBAL_USE:
                info(l10n("removing-use-flag", flag="-" + flag))
                GLOBAL_USE.remove(disableflag)

            set_config_value("USE", " ".join(sorted(GLOBAL_USE)))
            get_user_global_use.cache_clear()
        else:
            warning(l10n("flag-not-set-globally", flag=flag))
    get_use.cache_clear()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from typing import Dict, List, Optional, Set, cast

from .source import Source


class PhaseState:
    """
    Information passed to the phase functions for use during execution

    These fields match those of the same names listed and documented
    by :class:`~pybuild.Pybuild2`
    """

    A: List[Source]
    FILESDIR: str
    T: Optional[str]
    D: str
    USE: Set[str]
    WORKDIR: str
    ROOT: str
    UNFETCHED: List[Source]
    S: Optional[str]

    def __init__(self, build_dir: Optional[str] = None):
        if build_dir:
            self.T = os.path.join(build_dir, "temp")
        else:
            self.T = None

    @classmethod
    def from_json(cls, dictionary: Dict):
        new = PhaseState()
        for key, value in dictionary.items():
            setattr(new, key, value)
        if hasattr(new, "A"):
            new.A = [Source(**x) for x in cast(Dict, new.A)]
        if hasattr(new, "UNFETCHED"):
            new.UNFETCHED = [Source(**x) for x in cast(Dict, new.UNFETCHED)]
        return new



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Module for interacting with PORTMOD_CONFIG/repos.cfg"""

import ast
import configparser
import os
from logging import info, warning
from typing import Dict, List, Optional, Set

from portmodlib.l10n import l10n

from .config import get_config
from .globals import env
from .repo import RemoteRepo, Repo, get_repo_name
from .repo.metadata import get_archs, get_master_names


def _iterate_repos_cfg(path: str):
    """Parses contents of repos.cfg"""

    repo_config = configparser.ConfigParser()
    repo_config.read(path)

    for name, conf in repo_config.items():
        if name == "DEFAULT":
            # Ignore DEFAULT key, as it is always there. We will not use it
            continue

        yield name, conf


def parse_remote_repos(path: str) -> List[RemoteRepo]:
    repos = []
    for name, conf in _iterate_repos_cfg(path):
        repos.append(
            RemoteRepo(
                name,
                sync_type=conf.get("sync_type"),
                sync_uri=conf.get("sync_uri"),
                priority=int(conf.get("priority", "0")),
                description=conf.get("description"),
                quality=conf.get("quality"),
                arch=set(conf.get("arch", "").split(",")),
            )
        )

    return repos


def parse_repos(path: str) -> List[Repo]:
    repos = []
    for name, conf in _iterate_repos_cfg(path):
        if "location" not in conf:
            warning(l10n("repo-missing-location", repo=name))
            continue
        repo_name = name
        if os.path.exists(conf["location"]):
            name = get_repo_name(conf["location"])

        repos.append(
            Repo(
                name=repo_name,
                location=os.path.expanduser(conf["location"]),
                auto_sync=ast.literal_eval(conf.get("auto_sync", "False")),
                sync_type=conf.get("sync_type"),
                sync_uri=conf.get("sync_uri"),
                priority=int(conf.get("priority", "0")),
            )
        )
    return repos


def get_local_repos() -> Dict[str, Repo]:
    meta_repo = Repo(
        "meta",
        os.path.join(env.REPOS_DIR, "meta"),
        auto_sync=True,
        sync_type="git",
        sync_uri="https://gitlab.com/portmod/meta.git",
        priority=-1000,
    )
    repos = {"meta": meta_repo}
    if os.path.exists(env.REPOS_FILE):
        repos.update({repo.name: repo for repo in parse_repos(env.REPOS_FILE)})
        return repos
    return repos


def get_repos():
    """Returns available repositories"""
    added = {"meta"}
    local = get_local_repos()
    repos = [get_local_repos()["meta"]]

    def add_repos(to_add: Set[str]):
        nonlocal repos
        for repo_name in to_add:
            if repo_name not in local:
                warning(l10n("repo-does-not-exist", name=repo_name))
                continue
            repo = local[repo_name]
            if repo.name not in added:
                repos.append(repo)
                added.add(repo.name)
            add_repos(get_master_names(repo.location))

    add_repos(get_config()["REPOS"])

    # Sort repos by priority such that the highest priority appears first
    repos.sort(key=lambda x: (x.priority, x.name), reverse=True)

    return repos


def add_repo(repo: RemoteRepo) -> Optional[Repo]:
    """
    Adds repository to repos.cfg

    If the repository already exists, None will be returned
    Otherwise, the repo added to the file will be returned
    """
    # comment_prefixes="/" and allow_no_value makes comments be treated as keys and preserved.
    # Unfortunately comments on the line of the section header are not preserved
    repo_config = configparser.ConfigParser(comment_prefixes="/", allow_no_value=True)
    repo_config.read(env.REPOS_FILE)

    new_repo = None
    if repo.name not in repo_config:
        new_repo = Repo(
            repo.name,
            os.path.join(env.REPOS_DIR, repo.name),
            auto_sync=repo.sync_type == "git",
            sync_type=repo.sync_type,
            sync_uri=repo.sync_uri,
            priority=repo.priority or 0,
        )
        repo_config.add_section(repo.name)
        for key, value in new_repo.to_dict().items():
            repo_config.set(repo.name, key, str(value))

        info(l10n("repo-adding", name=repo.name, conf=env.REPOS_FILE))

        os.makedirs(os.path.dirname(env.REPOS_FILE), exist_ok=True)
        with open(env.REPOS_FILE, "w") as file:
            repo_config.write(file)
        env.REPOS.append(new_repo)

    return new_repo


def get_remote_repos(arch: Optional[str] = None) -> Dict[str, RemoteRepo]:
    """
    Returns repositories declared by the repositories currently known
    """
    # Note: This method should not be cached, as new declarations may be
    # added when synchronizing a repository, and may be immediately used by
    # a repository synchronizing afterwards.
    repos = {}
    for repo in env.REPOS:
        conf = os.path.join(repo.location, "metadata", "repos.cfg")
        if os.path.exists(conf):
            repos.update(
                {
                    remote_repo.name: remote_repo
                    for remote_repo in parse_remote_repos(conf)
                    if arch is None or arch in remote_repo.arch
                }
            )
    for repo in env.REPOS:
        if (
            repo.name not in repos
            and repo.name != "meta"
            and arch in get_archs(repo.location)
        ):
            repos[repo.name] = RemoteRepo(
                name=repo.name,
                sync_type=repo.sync_type,
                sync_uri=repo.sync_uri,
                priority=repo.priority,
                description="Local repository in repos.cfg",
                arch=get_archs(repo.location),
            )
    return repos



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os

import pytest

from portmod._cli.pybuild import pybuild_manifest
from portmod.globals import env
from portmod.repo import get_repo

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_manifest():
    """Tests that a manifests can be generated without a prefix set"""
    env.set_prefix(None)
    env.ALLOW_LOAD_ERROR = False
    path = os.path.join(
        get_repo("test").location,
        "test",
        "quill-of-feyfolken",
        "quill-of-feyfolken-2.0.2.pybuild",
    )
    pybuild_manifest(path)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import pytest

from portmodlib.atom import Atom, FQAtom, QualifiedAtom, Version, atom_sat, version_gt


def canimport(name: str) -> bool:
    """Returns true if the given module can be imported"""
    try:
        __import__(name)
        return True
    except ModuleNotFoundError:
        return False


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_atom(benchmark):
    """Test the speed of loading Atoms"""

    def test():
        Atom._CACHE = {}
        Atom("foo/bar-1.0::baz")

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_qualifed_atom(benchmark):
    """Test the speed of loading Atoms"""

    def test():
        Atom._CACHE = {}
        QualifiedAtom("foo/bar-1.0::baz")

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_fqatom(benchmark):
    """Test the speed of loading Atoms"""

    def test():
        Atom._CACHE = {}
        FQAtom("foo/bar-1.0::baz")

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_fqatom_cached(benchmark):
    """Test the speed of loading Atoms"""

    def test():
        FQAtom("foo/bar-1.0::baz")

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_cached(benchmark):
    """Test the speed of loading Atoms"""

    def test():
        Atom("foo/bar-1.0::baz")

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_atom_sat_similar(benchmark):
    """Test the speed of comparing similar Atoms"""

    def test():
        atom_sat(Atom("foo/bar-1.0.0"), Atom("~foo/bar-1.0-r1"))

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_atom_sat_different(benchmark):
    """Test the speed of comparing obviously different"""

    def test():
        atom_sat(Atom("foo/bar-1.0.0"), Atom("foo/baz-1.0-r1"))

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_version_gt(benchmark):
    """Test the speed of loading Atoms"""

    def test():
        version_gt("1.0.0", "1.0-r1")

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_version_comparison_simple(benchmark):
    """Test the speed of loading Atoms"""

    v1 = Version("1.0.0")
    v2 = Version("1.1")

    def test():
        v1 > v2

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_version_comparison_complex(benchmark):
    """Test the speed of loading Atoms"""

    v1 = Version("e1-1.0.0a_alpha2_beta3")
    v2 = Version("e1-1.0.0a_alpha2_beta3-r1")

    def test():
        v1 > v2

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_version_parse_simple(benchmark):
    """Test the speed of loading Atoms"""

    def test():
        Version("1.0.0")

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_version_parse_complex(benchmark):
    """Test the speed of loading Atoms"""

    def test():
        Version("e1-1.0.0a_alpha-r1")

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
def test_display_version(benchmark):
    """Test the speed of displaying versions"""

    x = Version("e1-1.0.0a_alpha-r1")

    def test():
        x.display()

    benchmark(test)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import pytest

from portmod.globals import env
from portmodlib._deprecated.vfs import find_file
from portmodlib.portmod import _get_hash


def canimport(name: str) -> bool:
    """Returns true if the given module can be imported"""
    try:
        __import__(name)
        return True
    except ModuleNotFoundError:
        return False


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
@pytest.mark.parametrize("alg", ["BLAKE2B", "MD5", "SHA512", "BLAKE3"])
@pytest.mark.parametrize(
    "buffer", [16 * 1024, 65536, 5 * 1024 * 1024, 512 * 1024 * 1024]
)
@pytest.mark.parametrize("file", ["TR_Data.bsa", "Quill of Feyfolken.omwaddon"])
def test_get_hash_rust(benchmark, alg, buffer, file):
    """Test the speed of loading Manifest files

    Note: requires an openmw configuration and a prefix named openmw
    """
    env.set_prefix("openmw")
    file = find_file(file)

    def test():
        _get_hash(file, [alg], buffer)

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
@pytest.mark.parametrize(
    "buffer", [16 * 1024, 65536, 5 * 1024 * 1024, 512 * 1024 * 1024]
)
@pytest.mark.parametrize("file", ["TR_Data.bsa", "Quill of Feyfolken.omwaddon"])
def test_get_hash_rust_multiple(benchmark, buffer, file):
    """Test the speed of loading Manifest files

    Note: requires an openmw configuration and a prefix named openmw
    """
    env.set_prefix("openmw")
    file = find_file(file)

    def test():
        _get_hash(file, ["BLAKE2B", "MD5", "SHA512", "BLAKE3"], buffer)

    benchmark(test)


@pytest.mark.skipif(
    not canimport("pytest_benchmark"),
    reason="requires pytest-benchmark",
)
@pytest.mark.parametrize(
    "buffer",
    [1024 * elem for elem in range(1, 10)]
    + [1024 * 1024 * elem for elem in range(1, 10)]
    + [1024 * 1024 * 10 * elem for elem in range(1, 10)],
)
@pytest.mark.parametrize("file", ["TR_Data.bsa", "Quill of Feyfolken.omwaddon"])
def test_get_hash_rust_blake3(benchmark, buffer, file):
    """Test the speed of loading Manifest files

    Note: requires an openmw configuration and a prefix named openmw
    """
    env.set_prefix("openmw")
    file = find_file(file)

    def test():
        _get_hash(file, ["BLAKE3"], buffer)

    benchmark(test)
